# SPDX-FileCopyrightText: 2021-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""With respect to camera frame and optics distortions, also export environment

with world, sky, atmospheric effects such as rainbows or smoke """

import bpy

import os
from imghdr import what  # imghdr is a python lib to identify image file types
from math import atan, pi, sqrt, degrees
from . import voxel_lib  # for smoke rendering
from .model_primitives import write_object_modifiers


# -------- find image texture # used for export_world -------- #


def image_format(img_f):
    """Identify input image filetypes to transmit to POV."""
    # First use the below explicit extensions to identify image file prospects
    ext = {
        "JPG": "jpeg",
        "JPEG": "jpeg",
        "GIF": "gif",
        "TGA": "tga",
        "IFF": "iff",
        "PPM": "ppm",
        "PNG": "png",
        "SYS": "sys",
        "TIFF": "tiff",
        "TIF": "tiff",
        "EXR": "exr",
        "HDR": "hdr",
    }.get(os.path.splitext(img_f)[-1].upper(), "")
    # Then, use imghdr to really identify the filetype as it can be different
    if not ext:
        # maybe add a check for if path exists here?
        print(" WARNING: texture image has no extension")  # too verbose

        ext = what(img_f)  # imghdr is a python lib to identify image file types
    return ext


def img_map(ts):
    """Translate mapping type from Blender UI to POV syntax and return that string."""
    image_map = ""
    texdata = bpy.data.textures[ts.texture]
    if ts.mapping == "FLAT":
        image_map = "map_type 0 "
    elif ts.mapping == "SPHERE":
        image_map = "map_type 1 "
    elif ts.mapping == "TUBE":
        image_map = "map_type 2 "

    # map_type 3 and 4 in development (?) (ENV in pov 3.8)
    # for POV-Ray, currently they just seem to default back to Flat (type 0)
    # elif ts.mapping=="?":
    #    image_map = " map_type 3 "
    # elif ts.mapping=="?":
    #    image_map = " map_type 4 "
    if ts.use_interpolation:  # Available if image sampling class reactivated?
        image_map += " interpolate 2 "
    if texdata.extension == "CLIP":
        image_map += " once "
    # image_map += "}"
    # if ts.mapping=='CUBE':
    #    image_map+= "warp { cubic } rotate <-90,0,180>"
    # no direct cube type mapping. Though this should work in POV 3.7
    # it doesn't give that good results(best suited to environment maps?)
    # if image_map == "":
    #    print(" No texture image  found ")
    return image_map


def img_map_transforms(ts):
    """Translate mapping transformations from Blender UI to POV syntax and return that string."""
    # XXX TODO: unchecked textures give error of variable referenced before assignment XXX
    # POV-Ray "scale" is not a number of repetitions factor, but ,its
    # inverse, a standard scale factor.
    # 0.5 Offset is needed relatively to scale because center of the
    # scale is 0.5,0.5 in blender and 0,0 in POV
    # Strange that the translation factor for scale is not the same as for
    # translate.
    # TODO: verify both matches with other blender renderers / internal in previous versions.
    image_map_transforms = ""
    image_map_transforms = "scale <%.4g,%.4g,%.4g> translate <%.4g,%.4g,%.4g>" % (
        ts.scale[0],
        ts.scale[1],
        ts.scale[2],
        ts.offset[0],
        ts.offset[1],
        ts.offset[2],
    )
    # image_map_transforms = (" translate <-0.5,-0.5,0.0> scale <%.4g,%.4g,%.4g> translate <%.4g,%.4g,%.4g>" % \
    # ( 1.0 / ts.scale.x,
    # 1.0 / ts.scale.y,
    # 1.0 / ts.scale.z,
    # (0.5 / ts.scale.x) + ts.offset.x,
    # (0.5 / ts.scale.y) + ts.offset.y,
    # ts.offset.z))
    # image_map_transforms = (
    # "translate <-0.5,-0.5,0> "
    # "scale <-1,-1,1> * <%.4g,%.4g,%.4g> "
    # "translate <0.5,0.5,0> + <%.4g,%.4g,%.4g>" % \
    # (1.0 / ts.scale.x,
    # 1.0 / ts.scale.y,
    # 1.0 / ts.scale.z,
    # ts.offset.x,
    # ts.offset.y,
    # ts.offset.z)
    # )
    return image_map_transforms


def img_map_bg(wts):
    """Translate world mapping from Blender UI to POV syntax and return that string."""
    tex = bpy.data.textures[wts.texture]
    image_mapBG = ""
    # texture_coords refers to the mapping of world textures:
    if wts.texture_coords in ["VIEW", "GLOBAL"]:
        image_mapBG = " map_type 0 "
    elif wts.texture_coords == "ANGMAP":
        image_mapBG = " map_type 1 "
    elif wts.texture_coords == "TUBE":
        image_mapBG = " map_type 2 "

    if tex.use_interpolation:
        image_mapBG += " interpolate 2 "
    if tex.extension == "CLIP":
        image_mapBG += " once "
    # image_mapBG += "}"
    # if wts.mapping == 'CUBE':
    #   image_mapBG += "warp { cubic } rotate <-90,0,180>"
    # no direct cube type mapping. Though this should work in POV 3.7
    # it doesn't give that good results(best suited to environment maps?)
    # if image_mapBG == "":
    #    print(" No background texture image  found ")
    return image_mapBG


def path_image(image):
    """Conform a path string to POV syntax to avoid POV errors."""
    return bpy.path.abspath(image.filepath, library=image.library).replace("\\", "/")
    # .replace("\\","/") to get only forward slashes as it's what POV prefers,
    # even on windows


# end find image texture
# -----------------------------------------------------------------------------


def export_camera(file, scene, global_matrix, render, tab_write):
    """Translate camera from Blender UI to POV syntax and write to exported file."""
    camera = scene.camera

    # DH disabled for now, this isn't the correct context
    active_object = None  # bpy.context.active_object # does not always work  MR
    matrix = global_matrix @ camera.matrix_world
    focal_point = camera.data.dof.focus_distance

    # compute resolution
    q_size = render.resolution_x / render.resolution_y
    tab_write(file, "#declare camLocation  = <%.6f, %.6f, %.6f>;\n" % matrix.translation[:])
    tab_write(
        file,
        (
            "#declare camLookAt = <%.6f, %.6f, %.6f>;\n"
            % tuple(degrees(e) for e in matrix.to_3x3().to_euler())
        ),
    )

    tab_write(file, "camera {\n")
    if scene.pov.baking_enable and active_object and active_object.type == "MESH":
        tab_write(file, "mesh_camera{ 1 3\n")  # distribution 3 is what we want here
        tab_write(file, "mesh{%s}\n" % active_object.name)
        tab_write(file, "}\n")
        tab_write(file, "location <0,0,.01>")
        tab_write(file, "direction <0,0,-1>")

    else:
        if camera.data.type == "ORTHO":
            # XXX todo: track when SensorHeightRatio was added to see if needed (not used)
            sensor_height_ratio = (
                render.resolution_x * camera.data.ortho_scale / render.resolution_y
            )
            tab_write(file, "orthographic\n")
            # Blender angle is radian so should be converted to degrees:
            # % (camera.data.angle * (180.0 / pi) )
            # but actually argument is not compulsory after angle in pov ortho mode
            tab_write(file, "angle\n")
            tab_write(file, "right <%6f, 0, 0>\n" % -camera.data.ortho_scale)
            tab_write(file, "location  <0, 0, 0>\n")
            tab_write(file, "look_at  <0, 0, -1>\n")
            tab_write(file, "up <0, %6f, 0>\n" % (camera.data.ortho_scale / q_size))

        elif camera.data.type == "PANO":
            tab_write(file, "panoramic\n")
            tab_write(file, "location  <0, 0, 0>\n")
            tab_write(file, "look_at  <0, 0, -1>\n")
            tab_write(file, "right <%s, 0, 0>\n" % -q_size)
            tab_write(file, "up <0, 1, 0>\n")
            tab_write(file, "angle  %f\n" % (360.0 * atan(16.0 / camera.data.lens) / pi))
        elif camera.data.type == "PERSP":
            # Standard camera otherwise would be default in pov
            tab_write(file, "location  <0, 0, 0>\n")
            tab_write(file, "look_at  <0, 0, -1>\n")
            tab_write(file, "right <%s, 0, 0>\n" % -q_size)
            tab_write(file, "up <0, 1, 0>\n")
            tab_write(
                file,
                "angle  %f\n"
                % (2 * atan(camera.data.sensor_width / 2 / camera.data.lens) * 180.0 / pi),
            )

        tab_write(
            file,
            "rotate  <%.6f, %.6f, %.6f>\n" % tuple(degrees(e) for e in matrix.to_3x3().to_euler()),
        )

        tab_write(file, "translate <%.6f, %.6f, %.6f>\n" % matrix.translation[:])
        if camera.data.dof.use_dof and (focal_point != 0 or camera.data.dof.focus_object):
            tab_write(
                file, "aperture %.3g\n" % (1 / (camera.data.dof.aperture_fstop * 10000) * 1000)
            )
            tab_write(
                file,
                "blur_samples %d %d\n"
                % (camera.data.pov.dof_samples_min, camera.data.pov.dof_samples_max),
            )
            tab_write(file, "variance 1/%d\n" % camera.data.pov.dof_variance)
            tab_write(file, "confidence %.3g\n" % camera.data.pov.dof_confidence)
            if camera.data.dof.focus_object:
                focal_ob = scene.objects[camera.data.dof.focus_object.name]
                matrix_blur = global_matrix @ focal_ob.matrix_world
                tab_write(file, "focal_point <%.4f,%.4f,%.4f>\n" % matrix_blur.translation[:])
            else:
                tab_write(file, "focal_point <0, 0, %f>\n" % focal_point)
    if camera.data.pov.normal_enable:
        tab_write(
            file,
            "normal {%s %.4f turbulence %.4f scale %.4f}\n"
            % (
                camera.data.pov.normal_patterns,
                camera.data.pov.cam_normal,
                camera.data.pov.turbulence,
                camera.data.pov.scale,
            ),
        )
    tab_write(file, "}\n")


exported_lights_count = 0


def export_lights(lamps, file, scene, global_matrix, tab_write):
    """Translate lights from Blender UI to POV syntax and write to exported file."""

    from .render import write_matrix, tab_write

    # Incremented after each lamp export to declare its target
    # currently used for Fresnel diffuse shader as their slope vector:
    global exported_lights_count
    # Get all lamps and keep their count in a global variable
    for exported_lights_count, ob in enumerate(lamps, start=1):
        lamp = ob.data

        matrix = global_matrix @ ob.matrix_world

        # Color is no longer modified by energy
        # any way to directly get bpy_prop_array as tuple?
        color = tuple(lamp.color)

        tab_write(file, "light_source {\n")
        tab_write(file, "< 0,0,0 >\n")
        tab_write(file, "color srgb<%.3g, %.3g, %.3g>\n" % color)

        if lamp.type == "POINT":
            pass
        elif lamp.type == "SPOT":
            tab_write(file, "spotlight\n")

            # Falloff is the main radius from the centre line
            tab_write(file, "falloff %.2f\n" % (degrees(lamp.spot_size) / 2.0))  # 1 TO 179 FOR BOTH
            tab_write(
                file, "radius %.6f\n" % ((degrees(lamp.spot_size) / 2.0) * (1.0 - lamp.spot_blend))
            )

            # Blender does not have a tightness equivalent, 0 is most like blender default.
            tab_write(file, "tightness 0\n")  # 0:10f

            tab_write(file, "point_at  <0, 0, -1>\n")
            if lamp.pov.use_halo:
                tab_write(file, "looks_like{\n")
                tab_write(file, "sphere{<0,0,0>,%.6f\n" % lamp.shadow_soft_size)
                tab_write(file, "hollow\n")
                tab_write(file, "material{\n")
                tab_write(file, "texture{\n")
                tab_write(file, "pigment{rgbf<1,1,1,%.4f>}\n" % (lamp.pov.halo_intensity * 5.0))
                tab_write(file, "}\n")
                tab_write(file, "interior{\n")
                tab_write(file, "media{\n")
                tab_write(file, "emission 1\n")
                tab_write(file, "scattering {1, 0.5}\n")
                tab_write(file, "density{\n")
                tab_write(file, "spherical\n")
                tab_write(file, "color_map{\n")
                tab_write(file, "[0.0 rgb <0,0,0>]\n")
                tab_write(file, "[0.5 rgb <1,1,1>]\n")
                tab_write(file, "[1.0 rgb <1,1,1>]\n")
                tab_write(file, "}\n")
                tab_write(file, "}\n")
                tab_write(file, "}\n")
                tab_write(file, "}\n")
                tab_write(file, "}\n")
                tab_write(file, "}\n")
                tab_write(file, "}\n")
        elif lamp.type == "SUN":
            tab_write(file, "parallel\n")
            tab_write(file, "point_at  <0, 0, -1>\n")  # *must* be after 'parallel'

        elif lamp.type == "AREA":
            # Area lights have no falloff type, so always use blenders lamp quad equivalent
            # for those?
            tab_write(file, "fade_power %d\n" % 2)
            size_x = lamp.size
            samples_x = lamp.pov.shadow_ray_samples_x
            if lamp.shape == "SQUARE":
                size_y = size_x
                samples_y = samples_x
            else:
                size_y = lamp.size_y
                samples_y = lamp.pov.shadow_ray_samples_y

            tab_write(
                file,
                "area_light <%.6f,0,0>,<0,%.6f,0> %d, %d\n"
                % (size_x, size_y, samples_x, samples_y),
            )
            tab_write(file, "area_illumination\n")
            if lamp.pov.shadow_ray_sample_method == "CONSTANT_JITTERED":
                if lamp.pov.use_jitter:
                    tab_write(file, "jitter\n")
            else:
                tab_write(file, "adaptive 1\n")
                tab_write(file, "jitter\n")

        # No shadow checked either at global or light level:
        if not scene.pov.use_shadows or (lamp.pov.shadow_method == "NOSHADOW"):
            tab_write(file, "shadowless\n")

        # Sun shouldn't be attenuated. Area lights have no falloff attribute so they
        # are put to type 2 attenuation a little higher above.
        if lamp.type not in {"SUN", "AREA"}:
            tab_write(file, "fade_power %d\n" % 2)  # Use blenders lamp quad equivalent

        write_matrix(file, matrix)

        tab_write(file, "}\n")

        # v(A,B) rotates vector A about origin by vector B.
        file.write(
            "#declare lampTarget%s= vrotate(<%.4g,%.4g,%.4g>,<%.4g,%.4g,%.4g>);\n"
            % (
                exported_lights_count,
                -ob.location.x,
                -ob.location.y,
                -ob.location.z,
                ob.rotation_euler.x,
                ob.rotation_euler.y,
                ob.rotation_euler.z,
            )
        )


def export_world(file, world, scene, global_matrix, tab_write):
    """write world as POV background and sky_sphere to exported file"""
    render = scene.pov
    agnosticrender = scene.render
    camera = scene.camera
    # matrix = global_matrix @ camera.matrix_world  # view dependant for later use NOT USED
    if not world:
        return

    # These lines added to get sky gradient (visible with PNG output)

    # For simple flat background:
    if not world.pov.use_sky_blend:
        # No alpha with Sky option:
        if render.alpha_mode == "SKY" and not agnosticrender.film_transparent:
            tab_write(
                file, "background {rgbt<%.3g, %.3g, %.3g, 0>}\n" % (world.pov.horizon_color[:])
            )

        elif render.alpha_mode == "STRAIGHT" or agnosticrender.film_transparent:
            tab_write(
                file, "background {rgbt<%.3g, %.3g, %.3g, 1>}\n" % (world.pov.horizon_color[:])
            )
        else:
            # Non fully transparent background could premultiply alpha and avoid
            # anti-aliasing display issue
            tab_write(
                file,
                "background {rgbft<%.3g, %.3g, %.3g, %.3g, 0>}\n"
                % (
                    world.pov.horizon_color[0],
                    world.pov.horizon_color[1],
                    world.pov.horizon_color[2],
                    render.alpha_filter,
                ),
            )

    world_tex_count = 0
    # For Background image textures
    for t in world.pov_texture_slots:  # risk to write several sky_spheres but maybe ok.
        if t:
            tex = bpy.data.textures[t.texture]
        if tex.type is not None:
            world_tex_count += 1
            # XXX No enable checkbox for world textures yet (report it?)
            # if t and tex.type == 'IMAGE' and t.use:
            if tex.type == "IMAGE":
                image_filename = path_image(tex.image)
                if tex.image.filepath != image_filename:
                    tex.image.filepath = image_filename
                if image_filename != "" and t.use_map_blend:
                    textures_blend = image_filename
                    # colvalue = t.default_value
                    t_blend = t

                # Commented below was an idea to make the Background image oriented as camera
                # taken here:
                # http://news.pov.org/pov.newusers/thread/%3Cweb.4a5cddf4e9c9822ba2f93e20@news.pov.org%3E/
                # Replace 4/3 by the ratio of each image found by some custom or existing
                # function
                # mapping_blend = (" translate <%.4g,%.4g,%.4g> rotate z*degrees" \
                #                "(atan((camLocation - camLookAt).x/(camLocation - " \
                #                "camLookAt).y)) rotate x*degrees(atan((camLocation - " \
                #                "camLookAt).y/(camLocation - camLookAt).z)) rotate y*" \
                #                "degrees(atan((camLocation - camLookAt).z/(camLocation - " \
                #                "camLookAt).x)) scale <%.4g,%.4g,%.4g>b" % \
                #                (t_blend.offset.x / 10 , t_blend.offset.y / 10 ,
                #                 t_blend.offset.z / 10, t_blend.scale.x ,
                #                 t_blend.scale.y , t_blend.scale.z))
                # using camera rotation valuesdirectly from blender seems much easier
                if t_blend.texture_coords == "ANGMAP":
                    mapping_blend = ""
                else:
                    # POV-Ray "scale" is not a number of repetitions factor, but its
                    # inverse, a standard scale factor.
                    # 0.5 Offset is needed relatively to scale because center of the
                    # UV scale is 0.5,0.5 in blender and 0,0 in POV
                    # Further Scale by 2 and translate by -1 are
                    # required for the sky_sphere not to repeat

                    mapping_blend = (
                        "scale 2 scale <%.4g,%.4g,%.4g> translate -1 "
                        "translate <%.4g,%.4g,%.4g> rotate<0,0,0> "
                        % (
                            (1.0 / t_blend.scale.x),
                            (1.0 / t_blend.scale.y),
                            (1.0 / t_blend.scale.z),
                            0.5 - (0.5 / t_blend.scale.x) - t_blend.offset.x,
                            0.5 - (0.5 / t_blend.scale.y) - t_blend.offset.y,
                            t_blend.offset.z,
                        )
                    )

                    # The initial position and rotation of the pov camera is probably creating
                    # the rotation offset should look into it someday but at least background
                    # won't rotate with the camera now.
                # Putting the map on a plane would not introduce the skysphere distortion and
                # allow for better image scale matching but also some waay to chose depth and
                # size of the plane relative to camera.
                tab_write(file, "sky_sphere {\n")
                tab_write(file, "pigment {\n")
                tab_write(
                    file,
                    'image_map{%s "%s" %s}\n'
                    % (image_format(textures_blend), textures_blend, img_map_bg(t_blend)),
                )
                tab_write(file, "}\n")
                tab_write(file, "%s\n" % mapping_blend)
                # The following layered pigment opacifies to black over the texture for
                # transmit below 1 or otherwise adds to itself
                tab_write(file, "pigment {rgb 0 transmit %s}\n" % tex.intensity)
                tab_write(file, "}\n")
                # tab_write(file, "scale 2\n")
                # tab_write(file, "translate -1\n")

    # For only Background gradient

    if world_tex_count == 0 and world.pov.use_sky_blend:
        tab_write(file, "sky_sphere {\n")
        tab_write(file, "pigment {\n")
        # maybe Should follow the advice of POV doc about replacing gradient
        # for skysphere..5.5
        tab_write(file, "gradient y\n")
        tab_write(file, "color_map {\n")

        if render.alpha_mode == "TRANSPARENT":
            tab_write(
                file,
                "[0.0 rgbft<%.3g, %.3g, %.3g, %.3g, 0>]\n"
                % (
                    world.pov.horizon_color[0],
                    world.pov.horizon_color[1],
                    world.pov.horizon_color[2],
                    render.alpha_filter,
                ),
            )
            tab_write(
                file,
                "[1.0 rgbft<%.3g, %.3g, %.3g, %.3g, 0>]\n"
                % (
                    world.pov.zenith_color[0],
                    world.pov.zenith_color[1],
                    world.pov.zenith_color[2],
                    render.alpha_filter,
                ),
            )
        if agnosticrender.film_transparent or render.alpha_mode == "STRAIGHT":
            tab_write(file, "[0.0 rgbt<%.3g, %.3g, %.3g, 0.99>]\n" % (world.pov.horizon_color[:]))
            # aa premult not solved with transmit 1
            tab_write(file, "[1.0 rgbt<%.3g, %.3g, %.3g, 0.99>]\n" % (world.pov.zenith_color[:]))
        else:
            tab_write(file, "[0.0 rgbt<%.3g, %.3g, %.3g, 0>]\n" % (world.pov.horizon_color[:]))
            tab_write(file, "[1.0 rgbt<%.3g, %.3g, %.3g, 0>]\n" % (world.pov.zenith_color[:]))
        tab_write(file, "}\n")
        tab_write(file, "}\n")
        tab_write(file, "}\n")
        # Sky_sphere alpha (transmit) is not translating into image alpha the same
        # way as 'background'

    # if world.pov.light_settings.use_indirect_light:
    #    scene.pov.radio_enable=1

    # Maybe change the above to a function copyInternalRenderer settings when
    # user pushes a button, then:
    # scene.pov.radio_enable = world.pov.light_settings.use_indirect_light
    # and other such translations but maybe this would not be allowed either?

    # -----------------------------------------------------------------------------

    mist = world.mist_settings

    if mist.use_mist:
        tab_write(file, "fog {\n")
        if mist.falloff == "LINEAR":
            tab_write(file, "distance %.6f\n" % ((mist.start + mist.depth) * 0.368))
        elif mist.falloff in ["QUADRATIC", "INVERSE_QUADRATIC"]:  # n**2 or squrt(n)?
            tab_write(file, "distance %.6f\n" % ((mist.start + mist.depth) ** 2 * 0.368))
        tab_write(
            file,
            "color rgbt<%.3g, %.3g, %.3g, %.3g>\n"
            % (*world.pov.horizon_color, (1.0 - mist.intensity)),
        )
        # tab_write(file, "fog_offset %.6f\n" % mist.start) #create a pov property to prepend
        # tab_write(file, "fog_alt %.6f\n" % mist.height) #XXX right?
        # tab_write(file, "turbulence 0.2\n")
        # tab_write(file, "turb_depth 0.3\n")
        tab_write(file, "fog_type 1\n")  # type2 for height
        tab_write(file, "}\n")
    if scene.pov.media_enable:
        tab_write(file, "media {\n")
        tab_write(
            file,
            "scattering { %d, rgb %.12f*<%.4g, %.4g, %.4g>\n"
            % (
                int(scene.pov.media_scattering_type),
                scene.pov.media_diffusion_scale,
                *(scene.pov.media_diffusion_color[:]),
            ),
        )
        if scene.pov.media_scattering_type == "5":
            tab_write(file, "eccentricity %.3g\n" % scene.pov.media_eccentricity)
        tab_write(file, "}\n")
        tab_write(
            file,
            "absorption %.12f*<%.4g, %.4g, %.4g>\n"
            % (scene.pov.media_absorption_scale, *(scene.pov.media_absorption_color[:])),
        )
        tab_write(file, "\n")
        tab_write(file, "samples %.d\n" % scene.pov.media_samples)
        tab_write(file, "}\n")


# -----------------------------------------------------------------------------
def export_rainbows(rainbows, file, scene, global_matrix, tab_write):
    """write all POV rainbows primitives to exported file"""

    from .render import write_matrix, tab_write

    pov_mat_name = "Default_texture"
    for ob in rainbows:
        povdataname = ob.data.name  # enough? XXX not used nor matrix fn?
        angle = degrees(ob.data.spot_size / 2.5)  # radians in blender (2
        width = ob.data.spot_blend * 10
        distance = ob.data.shadow_buffer_clip_start
        # eps=0.0000001
        # angle = br/(cr+eps) * 10 #eps is small epsilon variable to avoid dividing by zero
        # width = ob.dimensions[2] #now let's say width of rainbow is the actual proxy height
        # formerly:
        # cz-bz # let's say width of the rainbow is height of the cone (interfacing choice

        # v(A,B) rotates vector A about origin by vector B.
        # and avoid a 0 length vector by adding 1

        # file.write("#declare %s_Target= vrotate(<%.6g,%.6g,%.6g>,<%.4g,%.4g,%.4g>);\n" % \
        # (povdataname, -(ob.location.x+0.1), -(ob.location.y+0.1), -(ob.location.z+0.1),
        # ob.rotation_euler.x, ob.rotation_euler.y, ob.rotation_euler.z))

        direction = (  # XXX currently not used (replaced by track to?)
            ob.location.x,
            ob.location.y,
            ob.location.z,
        )  # not taking matrix into account
        rmatrix = global_matrix @ ob.matrix_world

        # ob.rotation_euler.to_matrix().to_4x4() * mathutils.Vector((0,0,1))
        # XXX Is result of the below offset by 90 degrees?
        up = ob.matrix_world.to_3x3()[1].xyz  # * global_matrix

        # XXX TO CHANGE:
        # formerly:
        # tab_write(file, "#declare %s = rainbow {\n"%povdataname)

        # clumsy for now but remove the rainbow from instancing
        # system because not an object. use lamps later instead of meshes

        # del data_ref[dataname]
        tab_write(file, "rainbow {\n")

        tab_write(file, "angle %.4f\n" % angle)
        tab_write(file, "width %.4f\n" % width)
        tab_write(file, "distance %.4f\n" % distance)
        tab_write(file, "arc_angle %.4f\n" % ob.pov.arc_angle)
        tab_write(file, "falloff_angle %.4f\n" % ob.pov.falloff_angle)
        tab_write(file, "direction <%.4f,%.4f,%.4f>\n" % rmatrix.translation[:])
        tab_write(file, "up <%.4f,%.4f,%.4f>\n" % (up[0], up[1], up[2]))
        tab_write(file, "color_map {\n")
        tab_write(file, "[0.000  color srgbt<1.0, 0.5, 1.0, 1.0>]\n")
        tab_write(file, "[0.130  color srgbt<0.5, 0.5, 1.0, 0.9>]\n")
        tab_write(file, "[0.298  color srgbt<0.2, 0.2, 1.0, 0.7>]\n")
        tab_write(file, "[0.412  color srgbt<0.2, 1.0, 1.0, 0.4>]\n")
        tab_write(file, "[0.526  color srgbt<0.2, 1.0, 0.2, 0.4>]\n")
        tab_write(file, "[0.640  color srgbt<1.0, 1.0, 0.2, 0.4>]\n")
        tab_write(file, "[0.754  color srgbt<1.0, 0.5, 0.2, 0.6>]\n")
        tab_write(file, "[0.900  color srgbt<1.0, 0.2, 0.2, 0.7>]\n")
        tab_write(file, "[1.000  color srgbt<1.0, 0.2, 0.2, 1.0>]\n")
        tab_write(file, "}\n")

        # tab_write(file, "texture {%s}\n"%pov_mat_name)
        write_object_modifiers(ob, file)
        # tab_write(file, "rotate x*90\n")
        # matrix = global_matrix @ ob.matrix_world
        # write_matrix(file, matrix)
        tab_write(file, "}\n")
        # continue #Don't render proxy mesh, skip to next object


def export_smoke(file, smoke_obj_name, smoke_path, comments, global_matrix):
    """export Blender smoke type fluids to pov media using df3 library"""

    from .render import write_matrix, tab_write

    flowtype = -1  # XXX todo: not used yet? should trigger emissive for fire type
    depsgraph = bpy.context.evaluated_depsgraph_get()
    smoke_obj = bpy.data.objects[smoke_obj_name].evaluated_get(depsgraph)
    domain = None
    smoke_modifier = None
    # Search smoke domain target for smoke modifiers
    for mod in smoke_obj.modifiers:
        if mod.type == "FLUID":
            if mod.fluid_type == "DOMAIN":
                domain = smoke_obj
                smoke_modifier = mod

            elif mod.fluid_type == "FLOW":
                if mod.flow_settings.flow_type == "BOTH":
                    flowtype = 2
                elif mod.flow_settings.flow_type == "FIRE":
                    flowtype = 1
                elif mod.flow_settings.flow_type == "SMOKE":
                    flowtype = 0
    eps = 0.000001  # XXX not used currently. restore from corner case ... zero div?
    if domain is not None:
        mod_set = smoke_modifier.domain_settings
        channeldata = []
        for v in mod_set.density_grid:
            channeldata.append(v.real)
            print(v.real)
        # -------- Usage in voxel texture:
        # channeldata = []
        # if channel == 'density':
        # for v in mod_set.density_grid:
        # channeldata.append(v.real)

        # if channel == 'fire':
        # for v in mod_set.flame_grid:
        # channeldata.append(v.real)

        resolution = mod_set.resolution_max
        big_res = [
            mod_set.domain_resolution[0],
            mod_set.domain_resolution[1],
            mod_set.domain_resolution[2],
        ]

        if mod_set.use_noise:
            big_res[0] = big_res[0] * (mod_set.noise_scale + 1)
            big_res[1] = big_res[1] * (mod_set.noise_scale + 1)
            big_res[2] = big_res[2] * (mod_set.noise_scale + 1)
        # else:
        # p = []
        # -------- gather smoke domain settings
        # BBox = domain.bound_box
        # p.append([BBox[0][0], BBox[0][1], BBox[0][2]])
        # p.append([BBox[6][0], BBox[6][1], BBox[6][2]])
        # mod_set = smoke_modifier.domain_settings
        # resolution = mod_set.resolution_max
        # smokecache = mod_set.point_cache
        # ret = read_cache(smokecache, mod_set.use_noise, mod_set.noise_scale + 1, flowtype)
        # res_x = ret[0]
        # res_y = ret[1]
        # res_z = ret[2]
        # density = ret[3]
        # fire = ret[4]

        # if res_x * res_y * res_z > 0:
        # -------- new cache format
        # big_res = []
        # big_res.append(res_x)
        # big_res.append(res_y)
        # big_res.append(res_z)
        # else:
        # max = domain.dimensions[0]
        # if (max - domain.dimensions[1]) < -eps:
        # max = domain.dimensions[1]

        # if (max - domain.dimensions[2]) < -eps:
        # max = domain.dimensions[2]

        # big_res = [int(round(resolution * domain.dimensions[0] / max, 0)),
        # int(round(resolution * domain.dimensions[1] / max, 0)),
        # int(round(resolution * domain.dimensions[2] / max, 0))]

        # if mod_set.use_noise:
        # big_res = [big_res[0] * (mod_set.noise_scale + 1),
        # big_res[1] * (mod_set.noise_scale + 1),
        # big_res[2] * (mod_set.noise_scale + 1)]

        # if channel == 'density':
        # channeldata = density

        # if channel == 'fire':
        # channeldata = fire

        # sc_fr = '%s/%s/%s/%05d' % (
        # efutil.export_path,
        # efutil.scene_filename(),
        # bpy.context.scene.name,
        # bpy.context.scene.frame_current
        # )
        #               if not os.path.exists( sc_fr ):
        #                   os.makedirs(sc_fr)
        #
        #               smoke_filename = '%s.smoke' % bpy.path.clean_name(domain.name)
        #               smoke_path = '/'.join([sc_fr, smoke_filename])
        #
        #               with open(smoke_path, 'wb') as smoke_file:
        #                   # Binary densitygrid file format
        #                   #
        #                   # File header
        #                   smoke_file.write(b'SMOKE')        #magic number
        #                   smoke_file.write(struct.pack('<I', big_res[0]))
        #                   smoke_file.write(struct.pack('<I', big_res[1]))
        #                   smoke_file.write(struct.pack('<I', big_res[2]))
        # Density data
        #                   smoke_file.write(struct.pack('<%df'%len(channeldata), *channeldata))
        #
        #               LuxLog('Binary SMOKE file written: %s' % (smoke_path))

        # return big_res[0], big_res[1], big_res[2], channeldata

        mydf3 = voxel_lib.df3(big_res[0], big_res[1], big_res[2])
        sim_sizeX, sim_sizeY, sim_sizeZ = mydf3.size()
        for x in range(sim_sizeX):
            for y in range(sim_sizeY):
                for z in range(sim_sizeZ):
                    mydf3.set(x, y, z, channeldata[((z * sim_sizeY + y) * sim_sizeX + x)])
        try:
            mydf3.exportDF3(smoke_path)
        except ZeroDivisionError:
            print("Show smoke simulation in 3D view before export")
        print("Binary smoke.df3 file written in preview directory")
        if comments:
            file.write("\n//--Smoke--\n\n")

        # Note: We start with a default unit cube.
        #       This is mandatory to read correctly df3 data - otherwise we could just directly use
        #       bbox coordinates from the start, and avoid scale/translate operations at the end...
        file.write("box{<0,0,0>, <1,1,1>\n")
        file.write("    pigment{ rgbt 1 }\n")
        file.write("    hollow\n")
        file.write("    interior{ //---------------------\n")
        file.write("        media{ method 3\n")
        file.write("               emission <1,1,1>*1\n")  # 0>1 for dark smoke to white vapour
        file.write("               scattering{ 1, // Type\n")
        file.write("                  <1,1,1>*0.1\n")
        file.write("                } // end scattering\n")
        file.write('                density{density_file df3 "%s"\n' % smoke_path)
        file.write("                        color_map {\n")
        file.write("                        [0.00 rgb 0]\n")
        file.write("                        [0.05 rgb 0]\n")
        file.write("                        [0.20 rgb 0.2]\n")
        file.write("                        [0.30 rgb 0.6]\n")
        file.write("                        [0.40 rgb 1]\n")
        file.write("                        [1.00 rgb 1]\n")
        file.write("                       } // end color_map\n")
        file.write("               } // end of density\n")
        file.write("               samples %i // higher = more precise\n" % resolution)
        file.write("         } // end of media --------------------------\n")
        file.write("    } // end of interior\n")

        # START OF TRANSFORMATIONS

        # Size to consider here are bbox dimensions (i.e. still in object space, *before* applying
        # loc/rot/scale and other transformations (like parent stuff), aka matrix_world).
        bbox = smoke_obj.bound_box
        dim = [
            abs(bbox[6][0] - bbox[0][0]),
            abs(bbox[6][1] - bbox[0][1]),
            abs(bbox[6][2] - bbox[0][2]),
        ]

        # We scale our cube to get its final size and shapes but still in *object* space (same as Blender's bbox).
        file.write("scale<%.6g,%.6g,%.6g>\n" % (dim[0], dim[1], dim[2]))

        # We offset our cube such that (0,0,0) coordinate matches Blender's object center.
        file.write("translate<%.6g,%.6g,%.6g>\n" % (bbox[0][0], bbox[0][1], bbox[0][2]))

        # We apply object's transformations to get final loc/rot/size in world space!
        # Note: we could combine the two previous transformations with this matrix directly...
        write_matrix(file, global_matrix @ smoke_obj.matrix_world)

        # END OF TRANSFORMATIONS

        file.write("}\n")

        # file.write("               interpolate 1\n")
        # file.write("               frequency 0\n")
        # file.write("   }\n")
        # file.write("}\n")



# SPDX-FileCopyrightText: 2021-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""User interface to camera frame, optics distortions, and environment

with world, sky, atmospheric effects such as rainbows or smoke """

import bpy
from bpy.utils import register_class, unregister_class
from bpy.types import Operator, Menu, Panel
from bl_operators.presets import AddPresetBase

from bl_ui import properties_data_camera

for member in dir(properties_data_camera):
    subclass = getattr(properties_data_camera, member)
    if hasattr(subclass, "COMPAT_ENGINES"):
        subclass.COMPAT_ENGINES.add("POVRAY_RENDER")
del properties_data_camera

# -------- Use only a subset of the world panels
# from bl_ui import properties_world

# # TORECREATE##DEPRECATED#properties_world.WORLD_PT_preview.COMPAT_ENGINES.add('POVRAY_RENDER')
# properties_world.WORLD_PT_context_world.COMPAT_ENGINES.add('POVRAY_RENDER')
# # TORECREATE##DEPRECATED#properties_world.WORLD_PT_world.COMPAT_ENGINES.add('POVRAY_RENDER')
# del properties_world

# -------- #
# Physics Main wrapping every class 'as is'
from bl_ui import properties_physics_common

for member in dir(properties_physics_common):
    subclass = getattr(properties_physics_common, member)
    if hasattr(subclass, "COMPAT_ENGINES"):
        subclass.COMPAT_ENGINES.add("POVRAY_RENDER")
del properties_physics_common

# Physics Rigid Bodies wrapping every class 'as is'
from bl_ui import properties_physics_rigidbody

for member in dir(properties_physics_rigidbody):
    subclass = getattr(properties_physics_rigidbody, member)
    if hasattr(subclass, "COMPAT_ENGINES"):
        subclass.COMPAT_ENGINES.add("POVRAY_RENDER")
del properties_physics_rigidbody

# Physics Rigid Body Constraint wrapping every class 'as is'
from bl_ui import properties_physics_rigidbody_constraint

for member in dir(properties_physics_rigidbody_constraint):
    subclass = getattr(properties_physics_rigidbody_constraint, member)
    if hasattr(subclass, "COMPAT_ENGINES"):
        subclass.COMPAT_ENGINES.add("POVRAY_RENDER")
del properties_physics_rigidbody_constraint

# Physics Smoke and fluids wrapping every class 'as is'
from bl_ui import properties_physics_fluid

for member in dir(properties_physics_fluid):
    subclass = getattr(properties_physics_fluid, member)
    if hasattr(subclass, "COMPAT_ENGINES"):
        subclass.COMPAT_ENGINES.add("POVRAY_RENDER")
del properties_physics_fluid

# Physics softbody wrapping every class 'as is'
from bl_ui import properties_physics_softbody

for member in dir(properties_physics_softbody):
    subclass = getattr(properties_physics_softbody, member)
    if hasattr(subclass, "COMPAT_ENGINES"):
        subclass.COMPAT_ENGINES.add("POVRAY_RENDER")
del properties_physics_softbody

# Physics Field wrapping every class 'as is'
from bl_ui import properties_physics_field

for member in dir(properties_physics_field):
    subclass = getattr(properties_physics_field, member)
    if hasattr(subclass, "COMPAT_ENGINES"):
        subclass.COMPAT_ENGINES.add("POVRAY_RENDER")
del properties_physics_field

# Physics Cloth wrapping every class 'as is'
from bl_ui import properties_physics_cloth

for member in dir(properties_physics_cloth):
    subclass = getattr(properties_physics_cloth, member)
    if hasattr(subclass, "COMPAT_ENGINES"):
        subclass.COMPAT_ENGINES.add("POVRAY_RENDER")
del properties_physics_cloth

# Physics Dynamic Paint wrapping every class 'as is'
from bl_ui import properties_physics_dynamicpaint

for member in dir(properties_physics_dynamicpaint):
    subclass = getattr(properties_physics_dynamicpaint, member)
    if hasattr(subclass, "COMPAT_ENGINES"):
        subclass.COMPAT_ENGINES.add("POVRAY_RENDER")
del properties_physics_dynamicpaint

from bl_ui import properties_particle

for member in dir(properties_particle):  # add all "particle" panels from blender
    subclass = getattr(properties_particle, member)
    if hasattr(subclass, "COMPAT_ENGINES"):
        subclass.COMPAT_ENGINES.add("POVRAY_RENDER")
del properties_particle


class CameraDataButtonsPanel:
    """Use this class to define buttons from the camera data tab of
    properties window."""

    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "data"
    COMPAT_ENGINES = {"POVRAY_RENDER"}

    @classmethod
    def poll(cls, context):
        cam = context.camera
        rd = context.scene.render
        return cam and (rd.engine in cls.COMPAT_ENGINES)


class WorldButtonsPanel:
    """Use this class to define buttons from the world tab of
    properties window."""

    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "world"
    COMPAT_ENGINES = {"POVRAY_RENDER"}

    @classmethod
    def poll(cls, context):
        wld = context.world
        rd = context.scene.render
        return wld and (rd.engine in cls.COMPAT_ENGINES)


# ---------------------------------------------------------------- #
# Camera Settings
# ---------------------------------------------------------------- #
class CAMERA_PT_POV_cam_dof(CameraDataButtonsPanel, Panel):
    """Use this class for camera depth of field focal blur buttons."""

    bl_label = "POV Aperture"
    COMPAT_ENGINES = {"POVRAY_RENDER"}
    bl_parent_id = "DATA_PT_camera_dof_aperture"
    bl_options = {"HIDE_HEADER"}
    # def draw_header(self, context):
    # cam = context.camera

    # self.layout.prop(cam.pov, "dof_enable", text="")

    def draw(self, context):
        layout = self.layout

        cam = context.camera

        layout.active = cam.dof.use_dof
        layout.use_property_split = True  # Active single-column layout

        flow = layout.grid_flow(
            row_major=True, columns=0, even_columns=True, even_rows=False, align=False
        )

        col = flow.column()
        col.label(text="F-Stop value will export as")
        col.label(text="POV aperture : " + "%.3f" % (1 / cam.dof.aperture_fstop * 1000))

        col = flow.column()
        col.prop(cam.pov, "dof_samples_min")
        col.prop(cam.pov, "dof_samples_max")
        col.prop(cam.pov, "dof_variance")
        col.prop(cam.pov, "dof_confidence")


class CAMERA_PT_POV_cam_nor(CameraDataButtonsPanel, Panel):
    """Use this class for camera normal perturbation buttons."""

    bl_label = "POV Perturbation"
    COMPAT_ENGINES = {"POVRAY_RENDER"}

    def draw_header(self, context):
        cam = context.camera

        self.layout.prop(cam.pov, "normal_enable", text="")

    def draw(self, context):
        layout = self.layout

        cam = context.camera

        layout.active = cam.pov.normal_enable

        layout.prop(cam.pov, "normal_patterns")
        layout.prop(cam.pov, "cam_normal")
        layout.prop(cam.pov, "turbulence")
        layout.prop(cam.pov, "scale")


class CAMERA_PT_POV_replacement_text(CameraDataButtonsPanel, Panel):
    """Use this class for camera text replacement field."""

    bl_label = "Custom POV Code"
    COMPAT_ENGINES = {"POVRAY_RENDER"}

    def draw(self, context):
        layout = self.layout

        cam = context.camera

        col = layout.column()
        col.label(text="Replace properties with:")
        col.prop(cam.pov, "replacement_text", text="")


# ---------------------------------------------------------------- #
# World background and sky sphere Settings
# ---------------------------------------------------------------- #


class WORLD_PT_POV_world(WorldButtonsPanel, Panel):
    """Use this class to define pov world buttons"""

    bl_label = "World"

    COMPAT_ENGINES = {"POVRAY_RENDER"}

    def draw(self, context):
        layout = self.layout

        world = context.world.pov

        row = layout.row(align=True)
        row.menu(WORLD_MT_POV_presets.__name__, text=WORLD_MT_POV_presets.bl_label)
        row.operator(WORLD_OT_POV_add_preset.bl_idname, text="", icon="ADD")
        row.operator(WORLD_OT_POV_add_preset.bl_idname, text="", icon="REMOVE").remove_active = True

        row = layout.row()
        row.prop(world, "use_sky_paper")
        row.prop(world, "use_sky_blend")
        row.prop(world, "use_sky_real")

        row = layout.row()
        row.column().prop(world, "horizon_color")
        col = row.column()
        col.prop(world, "zenith_color")
        col.active = world.use_sky_blend
        row.column().prop(world, "ambient_color")

        # row = layout.row()
        # row.prop(world, "exposure") #Re-implement later as a light multiplier
        # row.prop(world, "color_range")


class WORLD_PT_POV_mist(WorldButtonsPanel, Panel):
    """Use this class to define pov mist buttons."""

    bl_label = "Mist"
    bl_options = {"DEFAULT_CLOSED"}
    COMPAT_ENGINES = {"POVRAY_RENDER"}

    def draw_header(self, context):
        world = context.world

        self.layout.prop(world.mist_settings, "use_mist", text="")

    def draw(self, context):
        layout = self.layout

        world = context.world

        layout.active = world.mist_settings.use_mist

        split = layout.split()

        col = split.column()
        col.prop(world.mist_settings, "intensity")
        col.prop(world.mist_settings, "start")

        col = split.column()
        col.prop(world.mist_settings, "depth")
        col.prop(world.mist_settings, "height")

        layout.prop(world.mist_settings, "falloff")


class WORLD_MT_POV_presets(Menu):
    """Apply world preset to all concerned properties"""

    bl_label = "World Presets"
    preset_subdir = "pov/world"
    preset_operator = "script.execute_preset"
    draw = bpy.types.Menu.draw_preset


class WORLD_OT_POV_add_preset(AddPresetBase, Operator):
    """Add a World Preset recording current values"""

    bl_idname = "object.world_preset_add"
    bl_label = "Add World Preset"
    preset_menu = "WORLD_MT_POV_presets"

    # variable used for all preset values
    preset_defines = ["scene = bpy.context.scene"]

    # properties to store in the preset
    preset_values = [
        "scene.world.use_sky_blend",
        "scene.world.horizon_color",
        "scene.world.zenith_color",
        "scene.world.ambient_color",
        "scene.world.mist_settings.use_mist",
        "scene.world.mist_settings.intensity",
        "scene.world.mist_settings.depth",
        "scene.world.mist_settings.start",
        "scene.pov.media_enable",
        "scene.pov.media_scattering_type",
        "scene.pov.media_samples",
        "scene.pov.media_diffusion_scale",
        "scene.pov.media_diffusion_color",
        "scene.pov.media_absorption_scale",
        "scene.pov.media_absorption_color",
        "scene.pov.media_eccentricity",
    ]

    # where to store the preset
    preset_subdir = "pov/world"


class RENDER_PT_POV_media(WorldButtonsPanel, Panel):
    """Use this class to define a pov global atmospheric media buttons."""

    bl_label = "Atmosphere Media"
    COMPAT_ENGINES = {"POVRAY_RENDER"}

    def draw_header(self, context):
        scene = context.scene

        self.layout.prop(scene.pov, "media_enable", text="")

    def draw(self, context):
        layout = self.layout

        scene = context.scene

        layout.active = scene.pov.media_enable

        col = layout.column()
        col.prop(scene.pov, "media_scattering_type", text="")
        col = layout.column()
        col.prop(scene.pov, "media_samples", text="Samples")
        split = layout.split()
        col = split.column(align=True)
        col.label(text="Scattering:")
        col.prop(scene.pov, "media_diffusion_scale")
        col.prop(scene.pov, "media_diffusion_color", text="")
        col = split.column(align=True)
        col.label(text="Absorption:")
        col.prop(scene.pov, "media_absorption_scale")
        col.prop(scene.pov, "media_absorption_color", text="")
        if scene.pov.media_scattering_type == "5":
            col = layout.column()
            col.prop(scene.pov, "media_eccentricity", text="Eccentricity")


# ---------------------------------------------------------------- #
# Lights settings
# ---------------------------------------------------------------- #

# ----------------------------------------------------------------
# from bl_ui import properties_data_light
# for member in dir(properties_data_light):
# subclass = getattr(properties_data_light, member)
# try:
# subclass.COMPAT_ENGINES.add('POVRAY_RENDER')
# except BaseException as e:
# print e.__doc__
# print('An exception occurred: {}'.format(e))
# pass
# del properties_data_light
# -------- LIGHTS -------- #

from bl_ui import properties_data_light

# -------- These panels are kept
# properties_data_light.DATA_PT_custom_props_light.COMPAT_ENGINES.add('POVRAY_RENDER')
# properties_data_light.DATA_PT_context_light.COMPAT_ENGINES.add('POVRAY_RENDER')

# make some native panels contextual to some object variable
# by recreating custom panels inheriting their properties


class PovLightButtonsPanel(properties_data_light.DataButtonsPanel):
    """Use this class to define buttons from the light data tab of
    properties window."""

    COMPAT_ENGINES = {"POVRAY_RENDER"}
    POV_OBJECT_TYPES = {"RAINBOW"}

    @classmethod
    def poll(cls, context):
        obj = context.object
        # We use our parent class poll func too, avoids to re-define too much things...
        return (
            super(PovLightButtonsPanel, cls).poll(context)
            and obj
            and obj.pov.object_as not in cls.POV_OBJECT_TYPES
        )


# We cannot inherit from RNA classes (like e.g. properties_data_mesh.DATA_PT_vertex_groups).
# Complex py/bpy/rna interactions (with metaclass and all) simply do not allow it to work.
# So we simply have to explicitly copy here the interesting bits. ;)
from bl_ui import properties_data_light

# for member in dir(properties_data_light):
# subclass = getattr(properties_data_light, member)
# try:
# subclass.COMPAT_ENGINES.add('POVRAY_RENDER')
# except BaseException as e:
# print(e.__doc__)
# print('An exception occurred: {}'.format(e))
# pass

# Now only These panels are kept
properties_data_light.DATA_PT_custom_props_light.COMPAT_ENGINES.add("POVRAY_RENDER")
properties_data_light.DATA_PT_context_light.COMPAT_ENGINES.add("POVRAY_RENDER")


class LIGHT_PT_POV_preview(PovLightButtonsPanel, Panel):
    # XXX Needs update and docstring
    bl_label = properties_data_light.DATA_PT_preview.bl_label

    draw = properties_data_light.DATA_PT_preview.draw


class LIGHT_PT_POV_light(PovLightButtonsPanel, Panel):
    """UI panel to main pov light parameters"""

    # bl_label = properties_data_light.DATA_PT_light.bl_label

    # draw = properties_data_light.DATA_PT_light.draw
    # class DATA_PT_POV_light(DataButtonsPanel, Panel):
    bl_label = "Light"
    # COMPAT_ENGINES = {'POVRAY_RENDER'}

    def draw(self, context):
        layout = self.layout

        light = context.light

        layout.row().prop(light, "type", expand=True)

        split = layout.split()

        col = split.column()
        sub = col.column()
        sub.prop(light, "color", text="")
        sub.prop(light, "energy")

        if light.type in {"POINT", "SPOT"}:
            sub.prop(light, "shadow_soft_size", text="Radius")

        if light.type == "AREA":
            col.prop(light, "shape")

            sub = col.column(align=True)

            if light.shape in {'SQUARE', 'DISK'}:
                sub.prop(light, "size")
            elif light.shape in {'RECTANGLE', 'ELLIPSE'}:
                sub.prop(light, "size", text="Size X")
                sub.prop(light, "size_y", text="Y")

        # restore later as interface to POV light groups ?
        # col = split.column()
        # col.prop(light, "use_own_layer", text="This Layer Only")


class LIGHT_MT_POV_presets(Menu):
    """Use this class to define preset menu for pov lights."""

    bl_label = "Lamp Presets"
    preset_subdir = "pov/light"
    preset_operator = "script.execute_preset"
    draw = bpy.types.Menu.draw_preset


class LIGHT_OT_POV_add_preset(AddPresetBase, Operator):
    """Operator to add a Light Preset"""

    bl_idname = "object.light_preset_add"
    bl_label = "Add Light Preset"
    preset_menu = "LIGHT_MT_POV_presets"

    # variable used for all preset values
    preset_defines = ["lightdata = bpy.context.object.data"]

    # properties to store in the preset
    preset_values = ["lightdata.type", "lightdata.color"]

    # where to store the preset
    preset_subdir = "pov/light"


# Draw into the existing light panel
def light_panel_func(self, context):
    """Menu to browse and add light preset"""
    layout = self.layout

    row = layout.row(align=True)
    row.menu(LIGHT_MT_POV_presets.__name__, text=LIGHT_MT_POV_presets.bl_label)
    row.operator(LIGHT_OT_POV_add_preset.bl_idname, text="", icon="ADD")
    row.operator(LIGHT_OT_POV_add_preset.bl_idname, text="", icon="REMOVE").remove_active = True


"""#TORECREATE##DEPRECATED#
class LIGHT_PT_POV_sunsky(PovLightButtonsPanel, Panel):
    bl_label = properties_data_light.DATA_PT_sunsky.bl_label

    @classmethod
    def poll(cls, context):
        lamp = context.light
        engine = context.scene.render.engine
        return (lamp and lamp.type == 'SUN') and (engine in cls.COMPAT_ENGINES)

    draw = properties_data_light.DATA_PT_sunsky.draw

"""


class LIGHT_PT_POV_shadow(PovLightButtonsPanel, Panel):
    # Todo : update and docstring
    bl_label = "Shadow"

    @classmethod
    def poll(cls, context):
        light = context.light
        engine = context.scene.render.engine
        return light and (engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout

        light = context.light

        layout.row().prop(light.pov, "shadow_method", expand=True)

        split = layout.split()
        col = split.column()

        col.prop(light.pov, "use_halo")
        sub = col.column(align=True)
        sub.active = light.pov.use_halo
        sub.prop(light.pov, "halo_intensity", text="Intensity")

        if light.pov.shadow_method == "NOSHADOW" and light.type == "AREA":
            split = layout.split()

            col = split.column()
            col.label(text="Form factor sampling:")

            sub = col.row(align=True)

            if light.shape == "SQUARE":
                sub.prop(light, "shadow_ray_samples_x", text="Samples")
            elif light.shape == "RECTANGLE":
                sub.prop(light.pov, "shadow_ray_samples_x", text="Samples X")
                sub.prop(light.pov, "shadow_ray_samples_y", text="Samples Y")

        if light.pov.shadow_method != "NOSHADOW":
            split = layout.split()

            col = split.column()
            col.prop(light, "shadow_color", text="")

            # col = split.column()
            # col.prop(light.pov, "use_shadow_layer", text="This Layer Only")
            # col.prop(light.pov, "use_only_shadow")

        if light.pov.shadow_method == "RAY_SHADOW":
            split = layout.split()

            col = split.column()
            col.label(text="Sampling:")

            if light.type in {"POINT", "SUN", "SPOT"}:
                sub = col.row()

                sub.prop(light.pov, "shadow_ray_samples_x", text="Samples")
                # any equivalent in pov?
                # sub.prop(light, "shadow_soft_size", text="Soft Size")

            elif light.type == "AREA":
                sub = col.row(align=True)

                if light.shape == "SQUARE":
                    sub.prop(light.pov, "shadow_ray_samples_x", text="Samples")
                elif light.shape == "RECTANGLE":
                    sub.prop(light.pov, "shadow_ray_samples_x", text="Samples X")
                    sub.prop(light.pov, "shadow_ray_samples_y", text="Samples Y")


class LIGHT_PT_POV_spot(PovLightButtonsPanel, Panel):
    bl_label = properties_data_light.DATA_PT_spot.bl_label
    bl_parent_id = "LIGHT_PT_POV_light"
    bl_context = "data"

    @classmethod
    def poll(cls, context):
        lamp = context.light
        engine = context.scene.render.engine
        return (lamp and lamp.type == "SPOT") and (engine in cls.COMPAT_ENGINES)

    draw = properties_data_light.DATA_PT_spot.draw


class OBJECT_PT_POV_rainbow(PovLightButtonsPanel, Panel):
    """Use this class to define buttons from the rainbow panel of
    properties window. inheriting lamp buttons panel class"""

    bl_label = "POV-Ray Rainbow"
    COMPAT_ENGINES = {"POVRAY_RENDER"}
    # bl_options = {'HIDE_HEADER'}

    @classmethod
    def poll(cls, context):
        engine = context.scene.render.engine
        obj = context.object
        return obj and obj.pov.object_as == "RAINBOW" and (engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout

        obj = context.object

        col = layout.column()

        if obj.pov.object_as == "RAINBOW":
            if not obj.pov.unlock_parameters:
                col.prop(
                    obj.pov, "unlock_parameters", text="Exported parameters below", icon="LOCKED"
                )
                col.label(text="Rainbow projection angle: " + str(obj.data.spot_size))
                col.label(text="Rainbow width: " + str(obj.data.spot_blend))
                col.label(text="Rainbow distance: " + str(obj.data.shadow_buffer_clip_start))
                col.label(text="Rainbow arc angle: " + str(obj.pov.arc_angle))
                col.label(text="Rainbow falloff angle: " + str(obj.pov.falloff_angle))

            else:
                col.prop(
                    obj.pov, "unlock_parameters", text="Edit exported parameters", icon="UNLOCKED"
                )
                col.label(text="3D view proxy may get out of synch")
                col.active = obj.pov.unlock_parameters

                layout.operator("pov.cone_update", text="Update", icon="MESH_CONE")

                # col.label(text="Parameters:")
                col.prop(obj.data, "spot_size", text="Rainbow Projection Angle")
                col.prop(obj.data, "spot_blend", text="Rainbow width")
                col.prop(obj.data, "shadow_buffer_clip_start", text="Visibility distance")
                col.prop(obj.pov, "arc_angle")
                col.prop(obj.pov, "falloff_angle")


del properties_data_light


classes = (
    WORLD_PT_POV_world,
    WORLD_MT_POV_presets,
    WORLD_OT_POV_add_preset,
    WORLD_PT_POV_mist,
    RENDER_PT_POV_media,
    LIGHT_PT_POV_preview,
    LIGHT_PT_POV_light,
    LIGHT_PT_POV_shadow,
    LIGHT_PT_POV_spot,
    LIGHT_MT_POV_presets,
    LIGHT_OT_POV_add_preset,
    OBJECT_PT_POV_rainbow,
    CAMERA_PT_POV_cam_dof,
    CAMERA_PT_POV_cam_nor,
    CAMERA_PT_POV_replacement_text,
)


def register():

    for cls in classes:
        register_class(cls)
    LIGHT_PT_POV_light.prepend(light_panel_func)


def unregister():
    LIGHT_PT_POV_light.remove(light_panel_func)
    for cls in reversed(classes):
        unregister_class(cls)



# SPDX-FileCopyrightText: 2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

import pathlib
import pprint
import sys
import tempfile
import unittest
from pxr import Usd
from pxr import UsdUtils
from pxr import UsdGeom
from pxr import UsdShade
from pxr import Gf

import bpy

args = None


class AbstractUSDTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._tempdir = tempfile.TemporaryDirectory()
        cls.testdir = args.testdir
        cls.tempdir = pathlib.Path(cls._tempdir.name)

        return cls

    def setUp(self):
        self.assertTrue(
            self.testdir.exists(), "Test dir {0} should exist".format(self.testdir)
        )

    def tearDown(self):
        self._tempdir.cleanup()


class USDExportTest(AbstractUSDTest):
    def test_export_usdchecker(self):
        """Test exporting a scene and verifying it passes the usdchecker test suite"""
        bpy.ops.wm.open_mainfile(
            filepath=str(self.testdir / "usd_materials_export.blend")
        )
        export_path = self.tempdir / "usdchecker.usda"
        res = bpy.ops.wm.usd_export(
            filepath=str(export_path),
            export_materials=True,
            evaluation_mode="RENDER",
        )
        self.assertEqual({'FINISHED'}, res, f"Unable to export to {export_path}")

        checker = UsdUtils.ComplianceChecker(
            arkit=False,
            skipARKitRootLayerCheck=False,
            rootPackageOnly=False,
            skipVariants=False,
            verbose=False,
        )
        checker.CheckCompliance(str(export_path))

        failed_checks = {}

        # The ComplianceChecker does not know how to resolve <UDIM> tags, so
        # it will flag "textures/test_grid_<UDIM>.png" as a missing reference.
        # That reference is in fact OK, so we skip the rule for this test.
        to_skip = ("MissingReferenceChecker",)
        for rule in checker._rules:
            name = rule.__class__.__name__
            if name in to_skip:
                continue

            issues = rule.GetFailedChecks() + rule.GetWarnings() + rule.GetErrors()
            if not issues:
                continue

            failed_checks[name] = issues

        self.assertFalse(failed_checks, pprint.pformat(failed_checks))

    def compareVec3d(self, first, second):
        places = 5
        self.assertAlmostEqual(first[0], second[0], places)
        self.assertAlmostEqual(first[1], second[1], places)
        self.assertAlmostEqual(first[2], second[2], places)

    def test_export_extents(self):
        """Test that exported scenes contain have a properly authored extent attribute on each boundable prim"""
        bpy.ops.wm.open_mainfile(filepath=str(self.testdir / "usd_extent_test.blend"))
        export_path = self.tempdir / "usd_extent_test.usda"
        res = bpy.ops.wm.usd_export(
            filepath=str(export_path),
            export_materials=True,
            evaluation_mode="RENDER",
        )
        self.assertEqual({'FINISHED'}, res, f"Unable to export to {export_path}")

        # if prims are missing, the exporter must have skipped some objects
        stats = UsdUtils.ComputeUsdStageStats(str(export_path))
        self.assertEqual(stats["totalPrimCount"], 15, "Unexpected number of prims")

        # validate the overall world bounds of the scene
        stage = Usd.Stage.Open(str(export_path))
        scenePrim = stage.GetPrimAtPath("/scene")
        bboxcache = UsdGeom.BBoxCache(Usd.TimeCode.Default(), [UsdGeom.Tokens.default_])
        bounds = bboxcache.ComputeWorldBound(scenePrim)
        bound_min = bounds.GetRange().GetMin()
        bound_max = bounds.GetRange().GetMax()
        self.compareVec3d(bound_min, Gf.Vec3d(-5.752975881, -1, -2.798513651))
        self.compareVec3d(bound_max, Gf.Vec3d(1, 2.9515805244, 2.7985136508))

        # validate the locally authored extents
        prim = stage.GetPrimAtPath("/scene/BigCube/BigCubeMesh")
        extent = UsdGeom.Boundable(prim).GetExtentAttr().Get()
        self.compareVec3d(Gf.Vec3d(extent[0]), Gf.Vec3d(-1, -1, -2.7985137))
        self.compareVec3d(Gf.Vec3d(extent[1]), Gf.Vec3d(1, 1, 2.7985137))
        prim = stage.GetPrimAtPath("/scene/LittleCube/LittleCubeMesh")
        extent = UsdGeom.Boundable(prim).GetExtentAttr().Get()
        self.compareVec3d(Gf.Vec3d(extent[0]), Gf.Vec3d(-1, -1, -1))
        self.compareVec3d(Gf.Vec3d(extent[1]), Gf.Vec3d(1, 1, 1))
        prim = stage.GetPrimAtPath("/scene/Volume/Volume")
        extent = UsdGeom.Boundable(prim).GetExtentAttr().Get()
        self.compareVec3d(
            Gf.Vec3d(extent[0]), Gf.Vec3d(-0.7313742, -0.68043584, -0.5801515)
        )
        self.compareVec3d(
            Gf.Vec3d(extent[1]), Gf.Vec3d(0.7515701, 0.5500924, 0.9027928)
        )

    def test_opacity_threshold(self):
        # Note that the scene file used here is shared with a different test.
        # Here we assume that it has a Principled BSDF material with
        # a texture connected to its Base Color input.
        bpy.ops.wm.open_mainfile(filepath=str(self.testdir / "usd_materials_export.blend"))

        export_path = self.tempdir / "opaque_material.usda"
        res = bpy.ops.wm.usd_export(
            filepath=str(export_path),
            export_materials=True,
            evaluation_mode="RENDER",
        )
        self.assertEqual({'FINISHED'}, res, f"Unable to export to {export_path}")

        # Inspect and validate the exported USD for the opaque blend case.
        stage = Usd.Stage.Open(str(export_path))
        shader_prim = stage.GetPrimAtPath("/_materials/Material/Principled_BSDF")
        shader = UsdShade.Shader(shader_prim)
        opacity_input = shader.GetInput('opacity')
        self.assertEqual(opacity_input.HasConnectedSource(), False,
                         "Opacity input should not be connected for opaque material")
        self.assertAlmostEqual(opacity_input.Get(), 1.0, "Opacity input should be set to 1")

        # The material already has a texture input to the Base Color.
        # Now also link this texture to the Alpha input.
        # Set an opacity threshold appropriate for alpha clipping.
        mat = bpy.data.materials['Material']
        bsdf = mat.node_tree.nodes['Principled BSDF']
        tex_output = bsdf.inputs['Base Color'].links[0].from_node.outputs['Color']
        alpha_input = bsdf.inputs['Alpha']
        mat.node_tree.links.new(tex_output, alpha_input)
        bpy.data.materials['Material'].blend_method = 'CLIP'
        bpy.data.materials['Material'].alpha_threshold = 0.01
        export_path = self.tempdir / "alphaclip_material.usda"
        res = bpy.ops.wm.usd_export(
            filepath=str(export_path),
            export_materials=True,
            evaluation_mode="RENDER",
        )
        self.assertEqual({'FINISHED'}, res, f"Unable to export to {export_path}")

        # Inspect and validate the exported USD for the alpha clip case.
        stage = Usd.Stage.Open(str(export_path))
        shader_prim = stage.GetPrimAtPath("/_materials/Material/Principled_BSDF")
        shader = UsdShade.Shader(shader_prim)
        opacity_input = shader.GetInput('opacity')
        opacity_thres_input = shader.GetInput('opacityThreshold')
        self.assertEqual(opacity_input.HasConnectedSource(), True, "Alpha input should be connected")
        self.assertGreater(opacity_thres_input.Get(), 0.0, "Opacity threshold input should be > 0")

        # Modify material again, this time with alpha blend.
        bpy.data.materials['Material'].blend_method = 'BLEND'
        export_path = self.tempdir / "alphablend_material.usda"
        res = bpy.ops.wm.usd_export(
            filepath=str(export_path),
            export_materials=True,
            evaluation_mode="RENDER",
        )
        self.assertEqual({'FINISHED'}, res, f"Unable to export to {export_path}")

        # Inspect and validate the exported USD for the alpha blend case.
        stage = Usd.Stage.Open(str(export_path))
        shader_prim = stage.GetPrimAtPath("/_materials/Material/Principled_BSDF")
        shader = UsdShade.Shader(shader_prim)
        opacity_input = shader.GetInput('opacity')
        opacity_thres_input = shader.GetInput('opacityThreshold')
        self.assertEqual(opacity_input.HasConnectedSource(), True, "Alpha input should be connected")
        self.assertEqual(opacity_thres_input.Get(), None, "Opacity threshold should not be specified for alpha blend")


def main():
    global args
    import argparse

    if "--" in sys.argv:
        argv = [sys.argv[0]] + sys.argv[sys.argv.index("--") + 1:]
    else:
        argv = sys.argv

    parser = argparse.ArgumentParser()
    parser.add_argument("--testdir", required=True, type=pathlib.Path)
    args, remaining = parser.parse_known_args(argv)

    unittest.main(argv=remaining)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2016 Manfred Moitzi (mozman)
#
# SPDX-License-Identifier: MIT

# Purpose: entity classes, new implementation without dxf12/dxf13 layer
# Created: 17.04.2016

from __future__ import unicode_literals
__author__ = "mozman <mozman@gmx.at>"

import math

from mathutils import Vector

from . import const
from .color import TrueColor
from .styles import default_text_style
from .decode import decode

SPECIAL_CHARS = {
    'd': '°'
}


basic_attribs = {
    5: 'handle',
    6: 'linetype',
    8: 'layer',
    39: 'thickness',
    48: 'ltscale',
    62: 'color',
    67: 'paperspace',
    210: 'extrusion',
    284: 'shadow_mode',
    330: 'owner',
    370: 'line_weight',
    410: 'layout_tab_name',
}


class DXFEntity(object):
    def __init__(self):
        self.dxftype = 'ENTITY'
        self.handle = None
        self.owner = None
        self.paperspace = None
        self.layer = '0'
        self.linetype = None
        self.thickness = 0.0
        self.extrusion = None
        self.ltscale = 1.0
        self.line_weight = 0
        self.invisible = 0
        self.color = const.BYLAYER
        self.true_color = None
        self.transparency = None
        self.shadow_mode = None
        self.layout_tab_name = None

    def setup_attributes(self, tags):
        self.dxftype = tags.get_type()
        for code, value in tags.plain_tags():
            if code in basic_attribs:
                self.__setattr__(basic_attribs[code], value)
            elif code == 420:
                self.true_color = TrueColor(value)
            elif code == 440:
                self.transparency = 1. - float(value & 0xFF) / 255.
            else:
                yield code, value  # chain of generators

    def set_default_extrusion(self):  # call only for 2d entities with extrusion vector
        if self.extrusion is None:
            self.extrusion = (0., 0., 1.)

    def __str__(self):
        return "{} [{}]".format(self.dxftype, self.handle)


class Point(DXFEntity):
    def __init__(self):
        super(Point, self).__init__()
        self.point = (0, 0, 0)

    def setup_attributes(self, tags):
        for code, value in super(Point, self).setup_attributes(tags):
            if code == 10:
                self.point = value
            else:
                yield code, value  # chain of generators
        self.set_default_extrusion()


class Line(DXFEntity):
    def __init__(self):
        super(Line, self).__init__()
        self.start = (0, 0, 0)
        self.end = (0, 0, 0)

    def setup_attributes(self, tags):
        for code, value in super(Line, self).setup_attributes(tags):
            if code == 10:
                self.start = value
            elif code == 11:
                self.end = value
            else:
                yield code, value  # chain of generators


class Circle(DXFEntity):
    def __init__(self):
        super(Circle, self).__init__()
        self.center = (0, 0, 0)
        self.radius = 1.0

    def setup_attributes(self, tags):
        for code, value in super(Circle, self).setup_attributes(tags):
            if code == 10:
                self.center = value
            elif code == 40:
                self.radius = value
            else:
                yield code, value  # chain of generators
        self.set_default_extrusion()


class Arc(Circle):
    def __init__(self):
        super(Arc, self).__init__()
        self.start_angle = 0.
        self.end_angle = 360.

    def setup_attributes(self, tags):
        for code, value in super(Arc, self).setup_attributes(tags):
            if code == 50:
                self.start_angle = value
            elif code == 51:
                self.end_angle = value
            else:
                yield code, value  # chain of generators
        self.set_default_extrusion()

TRACE_CODES = frozenset((10, 11, 12, 13))


class Trace(DXFEntity):
    def __init__(self):
        super(Trace, self).__init__()
        self.points = []

    def setup_attributes(self, tags):
        for code, value in super(Trace, self).setup_attributes(tags):
            if code in TRACE_CODES:
                self.points.append(value)
            else:
                yield code, value  # chain of generators
        self.set_default_extrusion()


Solid = Trace


class Face(Trace):
    def __init__(self):
        super(Face, self).__init__()
        self.points = []
        self.invisible_edge = 0

    def setup_attributes(self, tags):
        for code, value in super(Face, self).setup_attributes(tags):
            if code == 70:
                self.invisible_edge = value
            else:
                yield code, value  # chain of generators
        self.set_default_extrusion()

    def is_edge_invisible(self, edge):
        # edges 0 .. 3
        return bool(self.invisible_edge & (1 << edge))


class Text(DXFEntity):
    def __init__(self):
        super(Text, self).__init__()
        self.insert = (0., 0.)
        self.height = 1.0
        self.text = ""
        self.rotation = 0.
        self.oblique = 0.
        self.style = "STANDARD"
        self.width = 1.
        self.is_backwards = False
        self.is_upside_down = False
        self.halign = 0
        self.valign = 0
        self.align_point = None
        self.font = None
        self.big_font = None

    def setup_attributes(self, tags):
        for code, value in super(Text, self).setup_attributes(tags):
            if code == 10:
                self.insert = value
            elif code == 11:
                self.align_point = value
            elif code == 1:
                self.text = value
            elif code == 7:
                self.style = value
            elif code == 40:
                self.height = value
            elif code == 41:
                self.width = value
            elif code == 50:
                self.rotation = value
            elif code == 51:
                self.oblique = value
            elif code == 71:
                self.is_backwards = bool(value & 2)
                self.is_upside_down = bool(value & 4)
            elif code == 72:
                self.halign = value
            elif code == 73:
                self.valign = value
            else:
                yield code, value  # chain of generators
        self.set_default_extrusion()

    def resolve_text_style(self, text_styles):
        style = text_styles.get(self.style, None)
        if style is None:
            style = default_text_style
        if self.height == 0:
            self.height = style.height
        if self.width == 0:
            self.width = style.width
        if self.oblique is None:
            self.oblique = style.oblique
        if self.is_backwards is None:
            self.is_backwards = style.is_backwards
        if self.is_upside_down is None:
            self.is_upside_down = style.is_upside_down
        if self.font is None:
            self.font = style.font
        if self.big_font is None:
            self.big_font = style.big_font

    def plain_text(self):
        chars = []
        raw_chars = list(reversed(self.text))  # text splitted into chars, in reversed order for efficient pop()
        while len(raw_chars):
            char = raw_chars.pop()
            if char == '%':  # formatting codes and special characters
                if len(raw_chars) and raw_chars[-1] == '%':
                    raw_chars.pop()  # '%'
                    if len(raw_chars):
                        special_char = raw_chars.pop()  # command char
                        chars.append(SPECIAL_CHARS.get(special_char, ""))
                else:  # char is just a single '%'
                    chars.append(char)
            else:  # char is what it is, a character
                chars.append(char)
        return "".join(chars)


class Attrib(Text):
    def __init__(self):
        super(Attrib, self).__init__()
        self.field_length = 0
        self.tag = ""

    def setup_attributes(self, tags):
        for code, value in super(Attrib, self).setup_attributes(tags):
            if code == 2:
                self.tag = value
            elif code == 73:
                self.field_length = value
            else:
                yield code, value


class Insert(DXFEntity):
    def __init__(self):
        super(Insert, self).__init__()
        self.name = ""
        self.insert = (0., 0., 0.)
        self.rotation = 0.
        self.scale = (1., 1., 1.)
        self.row_count = 1
        self.row_spacing = 0.
        self.col_count = 1
        self.col_spacing = 0.
        self.attribsfollow = False
        self.attribs = []

    def setup_attributes(self, tags):
        xscale = 1.
        yscale = 1.
        zscale = 1.
        for code, value in super(Insert, self).setup_attributes(tags):
            if code == 2:
                self.name = value
            elif code == 10:
                self.insert = value
            elif code == 41:
                xscale = value
            elif code == 42:
                yscale = value
            elif code == 43:
                zscale = value
            elif code == 44:
                self.col_spacing = value
            elif code == 45:
                self.row_spacing = value
            elif code == 50:
                self.rotation = value
            elif code == 66:
                self.attribsfollow = bool(value)
            elif code == 70:
                self.col_count = value
            elif code == 71:
                self.row_count = value
            else:
                yield code, value  # chain of generators
        self.scale = (xscale, yscale, zscale)
        self.set_default_extrusion()

    def find_attrib(self, attrib_tag):
        for attrib in self.attribs:
            if attrib.tag == attrib_tag:
                return attrib
        return None

    def append_data(self, attribs):
        self.attribs = attribs


class Polyline(DXFEntity):
    LINE_TYPES = frozenset(('spline2d', 'polyline2d', 'polyline3d'))

    def __init__(self):
        super(Polyline, self).__init__()
        self.vertices = []  # set in append data
        self.points = []  # set in append data
        self.control_points = []  # set in append data
        self.width = []  # set in append data
        self.bulge = []  # set in append data
        self.tangents = []  # set in append data
        self.flags = 0
        self.mode = 'polyline2d'
        self.mcount = 0
        self.ncount = 0
        self.default_start_width = 0.
        self.default_end_width = 0.
        self.is_mclosed = False
        self.is_nclosed = False
        self.is_closed = False
        self.elevation = (0., 0., 0.)
        self.m_smooth_density = 0.
        self.n_smooth_density = 0.
        self.smooth_type = 0
        self.spline_type = None

    def setup_attributes(self, tags):
        def get_mode():
            flags = self.flags
            if flags & const.POLYLINE_SPLINE_FIT_VERTICES_ADDED:
                return 'spline2d'
            elif flags & const.POLYLINE_3D_POLYLINE:
                return 'polyline3d'
            elif flags & const.POLYLINE_3D_POLYMESH:
                return 'polymesh'
            elif flags & const.POLYLINE_POLYFACE:
                return 'polyface'
            else:
                return 'polyline2d'

        for code, value in super(Polyline, self).setup_attributes(tags):
            if code == 10:
                self.elevation = value
            elif code == 40:
                self.default_start_width = value
            elif code == 41:
                self.default_end_width = value
            elif code == 70:
                self.flags = value
            elif code == 71:
                self.mcount = value
            elif code == 72:
                self.ncount = value
            elif code == 73:
                self.m_smooth_density = value
            elif code == 73:
                self.n_smooth_density = value
            elif code == 75:
                self.smooth_type = value
            else:
                yield code, value  # chain of generators

        self.mode = get_mode()
        if self.mode == 'spline2d':
            if self.smooth_type == const.POLYMESH_CUBIC_BSPLINE:
                self.spline_type = 'cubic_bspline'
            elif self.smooth_type == const.POLYMESH_QUADRIC_BSPLINE:
                self.spline_type = 'quadratic_bspline'
            elif self.smooth_type == const.POLYMESH_BEZIER_SURFACE:
                self.spline_type = 'bezier_curve'  # is this a valid spline type for DXF12?
        self.is_mclosed = bool(self.flags & const.POLYLINE_MESH_CLOSED_M_DIRECTION)
        self.is_nclosed = bool(self.flags & const.POLYLINE_MESH_CLOSED_N_DIRECTION)
        self.is_closed = self.is_mclosed
        self.set_default_extrusion()

    def __len__(self):
        return len(self.vertices)

    def __getitem__(self, item):
        return self.vertices[item]

    def __iter__(self):
        return iter(self.vertices)

    def append_data(self, vertices):
        def default_width(start_width, end_width):
            if start_width == 0.:
                start_width = self.default_start_width
            if end_width == 0.:
                end_width = self.default_end_width
            return start_width, end_width

        self.vertices = vertices
        if self.mode in Polyline.LINE_TYPES:
            for vertex in self.vertices:
                if vertex.flags & const.VTX_SPLINE_FRAME_CONTROL_POINT:
                    self.control_points.append(vertex.location)
                else:
                    self.points.append(vertex.location)
                    self.width.append(default_width(vertex.start_width, vertex.end_width))
                    self.bulge.append(vertex.bulge)
                    self.tangents.append(vertex.tangent if vertex.flags & const.VTX_CURVE_FIT_TANGENT else None)

    def cast(self):
        if self.mode == 'polyface':
            return PolyFace(self)
        elif self.mode == 'polymesh':
            return PolyMesh(self)
        else:
            return self


class SubFace(object):
    def __init__(self, face_record, vertices):
        self._vertices = vertices
        self.face_record = face_record

    def __len__(self):
        return len(self.face_record.vtx)

    def __getitem__(self, item):
        return self._vertices[self._vertex_index(item)]

    def __iter__(self):
        return (self._vertices[index].location for index in self.indices())

    def _vertex_index(self, pos):
        return abs(self.face_record.vtx[pos]) - 1

    def indices(self):
        return tuple(abs(i)-1 for i in self.face_record.vtx if i != 0)

    def is_edge_visible(self, pos):
        return self.face_record.vtx[pos] > 0


class PolyShape(object):
    def __init__(self, polyline, dxftype):
        # copy all dxf attributes from polyline
        for key, value in polyline.__dict__.items():
            self.__dict__[key] = value
        self.dxftype = dxftype

    def __str__(self):
        return "{} [{}]".format(self.dxftype, self.handle)


class PolyFace(PolyShape):
    def __init__(self, polyline):
        VERTEX_FLAGS = const.VTX_3D_POLYFACE_MESH_VERTEX + const.VTX_3D_POLYGON_MESH_VERTEX

        def is_vertex(flags):
            return flags & VERTEX_FLAGS == VERTEX_FLAGS

        super(PolyFace, self).__init__(polyline, 'POLYFACE')
        vertices = []
        face_records = []
        for vertex in polyline.vertices:
            (vertices if is_vertex(vertex.flags) else face_records).append(vertex)

        self._face_records = face_records

    def __getitem__(self, item):
        return SubFace(self._face_records[item], self.vertices)

    def __len__(self):
        return len(self._face_records)

    def __iter__(self):
        return (SubFace(f, self.vertices) for f in self._face_records)


class PolyMesh(PolyShape):
    def __init__(self, polyline):
        super(PolyMesh, self).__init__(polyline, 'POLYMESH')

    def __iter__(self):
        return iter(self.vertices)

    def get_location(self, pos):
        return self.get_vertex(pos).location

    def get_vertex(self, pos):
        m, n = pos
        if 0 <= m < self.mcount and 0 <= n < self.ncount:
            pos = m * self.ncount + n
            return self.vertices[pos]
        else:
            raise IndexError(repr(pos))


class Vertex(DXFEntity):
    def __init__(self):
        super(Vertex, self).__init__()
        self.location = (0., 0., 0.)
        self.flags = 0
        self.start_width = 0.
        self.end_width = 0.
        self.bulge = 0.
        self.tangent = None
        self.vtx = None

    def setup_attributes(self, tags):
        vtx0 = 0
        vtx1 = 0
        vtx2 = 0
        vtx3 = 0
        for code, value in super(Vertex, self).setup_attributes(tags):
            if code == 10:
                self.location = value
            elif code == 40:
                self.start_width = value
            elif code == 41:
                self.end_width = value
            elif code == 42:
                self.bulge = value
            elif code == 50:
                self.tangent = value
            elif code == 70:
                self.flags = value
            elif code == 71:
                vtx0 = value
            elif code == 72:
                vtx1 = value
            elif code == 73:
                vtx2 = value
            elif code == 74:
                vtx3 = value
            else:
                yield code, value  # chain of generators
        indices = (vtx0, vtx1, vtx2, vtx3)
        if any(indices):
            self.vtx = indices

    def __getitem__(self, item):
        return self.location[item]

    def __iter__(self):
        return iter(self.location)


class Block(DXFEntity):
    def __init__(self):
        super(Block, self).__init__()
        self.basepoint = (0, 0, 0)
        self.name = ''
        self.description = ''
        self.flags = 0
        self.xrefpath = ""
        self._entities = []

    def setup_attributes(self, tags):
        for code, value in super(Block, self).setup_attributes(tags):
            if code == 2:
                self.name = value
            elif code == 4:
                self.description = value
            elif code == 1:
                self.xrefpath = value
            elif code == 10:
                self.basepoint = value
            elif code == 70:
                self.flags = value
            else:
                yield code, value  # chain of generators

    @property
    def is_xref(self):
        return bool(self.flags & const.BLK_XREF)

    @property
    def is_xref_overlay(self):
        return bool(self.flags & const.BLK_XREF_OVERLAY)

    @property
    def is_anonymous(self):
        return bool(self.flags & const.BLK_ANONYMOUS)

    def set_entities(self, entities):
        self._entities = entities

    def __iter__(self):
        return iter(self._entities)

    def __getitem__(self, item):
        return self._entities[item]

    def __len__(self):
        return len(self._entities)


class LWPolyline(DXFEntity):
    def __init__(self):
        super(LWPolyline, self).__init__()
        self.points = []
        self.width = []
        self.bulge = []
        self.elevation = 0.
        self.const_width = 0.
        self.flags = 0

    def setup_attributes(self, tags):
        bulge, start_width, end_width = 0., 0., 0.
        init = True

        for code, value in super(LWPolyline, self).setup_attributes(tags):
            if code == 10:
                if not init:
                    self.bulge.append(bulge)
                    self.width.append((start_width, end_width))
                    bulge, start_width, end_width = 0., 0., 0.
                self.points.append(value)
                init = False
            elif code == 40:
                start_width = value
            elif code == 41:
                end_width = value
            elif code == 42:
                bulge = value
            elif code == 38:
                self.elevation = value
            elif code == 39:
                self.thickness = value
            elif code == 43:
                self.const_width = value
            elif code == 70:
                self.flags = value
            elif code == 210:
                self.extrusion = value
            else:
                yield code, value  # chain of generators

        # add values for the last point
        self.bulge.append(bulge)
        self.width.append((start_width, end_width))

        if self.const_width != 0.:
            self.width = []
        self.set_default_extrusion()

    @property
    def is_closed(self):
        return bool(self.flags & 1)

    def __len__(self):
        return len(self.points)

    def __getitem__(self, item):
        return self.points[item]

    def __iter__(self):
        return iter(self.points)


class Ellipse(DXFEntity):
    def __init__(self):
        super(Ellipse, self).__init__()
        self.center = (0., 0., 0.)
        self.major_axis = (1., 0., 0.)
        self.ratio = 1.0
        self.start_param = 0.
        self.end_param = 6.283185307179586

    def setup_attributes(self, tags):
        for code, value in super(Ellipse, self).setup_attributes(tags):
            if code == 10:
                self.center = value
            elif code == 11:
                self.major_axis = value
            elif code == 40:
                self.ratio = value
            elif code == 41:
                self.start_param = value
            elif code == 42:
                self.end_param = value
            else:
                yield code, value  # chain of generators
        self.set_default_extrusion()


class Ray(DXFEntity):
    def __init__(self):
        super(Ray, self).__init__()
        self.start = (0, 0, 0)
        self.unit_vector = (1, 0, 0)

    def setup_attributes(self, tags):
        for code, value in super(Ray, self).setup_attributes(tags):
            if code == 10:
                self.start = value
            elif code == 11:
                self.unit_vector = value
            else:
                yield code, value  # chain of generators


def deg2vec(deg):
    rad = float(deg) * math.pi / 180.0
    return math.cos(rad), math.sin(rad), 0.


##################################################
# MTEXT inline codes
# \L    Start underline
# \l    Stop underline
# \O    Start overstrike
# \o    Stop overstrike
# \K    Start strike-through
# \k    Stop strike-through
# \P    New paragraph (new line)
# \pxi  Control codes for bullets, numbered paragraphs and columns
# \X    Paragraph wrap on the dimension line (only in dimensions)
# \Q    Slanting (obliquing) text by angle - e.g. \Q30;
# \H    Text height - e.g. \H3x;
# \W    Text width - e.g. \W0.8x;
# \F    Font selection
#
#     e.g. \Fgdt;o - GDT-tolerance
#     e.g. \Fkroeger|b0|i0|c238|p10 - font Kroeger, non-bold, non-italic, codepage 238, pitch 10
#
# \S    Stacking, fractions
#
#     e.g. \SA^B:
#     A
#     B
#     e.g. \SX/Y:
#     X
#     -
#     Y
#     e.g. \S1#4:
#     1/4
#
# \A    Alignment
#
#     \A0; = bottom
#     \A1; = center
#     \A2; = top
#
# \C    Color change
#
#     \C1; = red
#     \C2; = yellow
#     \C3; = green
#     \C4; = cyan
#     \C5; = blue
#     \C6; = magenta
#     \C7; = white
#
# \T    Tracking, char.spacing - e.g. \T2;
# \~    Non-wrapping space, hard space
# {}    Braces - define the text area influenced by the code
# \ Escape character - e.g. \\ = "\", \{ = "{"
#
# Codes and braces can be nested up to 8 levels deep

ESCAPED_CHARS = "\\{}"
GROUP_CHARS = "{}"
ONE_CHAR_COMMANDS = "PLlOoKkX"


class MText(DXFEntity):
    def __init__(self):
        super(MText, self).__init__()
        self.insert = (0., 0., 0.)
        self.raw_text = ""
        self.height = 0.
        self.rect_width = None
        self.horizontal_width = None
        self.vertical_height = None
        self.line_spacing = 1.
        self.attachment_point = 1
        self.style = 'STANDARD'
        self.xdirection = (1., 0., 0.)
        self.font = None
        self.big_font = None

    def setup_attributes(self, tags):
        text = ""
        lines = []
        rotation = 0.
        xdir = None
        for code, value in super(MText, self).setup_attributes(tags):
            if code == 10:
                self.insert = value
            elif code == 11:
                xdir = value
            elif code == 1:
                text = value
            elif code == 3:
                lines.append(value)
            elif code == 7:
                self.style = value
            elif code == 40:
                self.height = value
            elif code == 41:
                self.rect_width = value
            elif code == 42:
                self.horizontal_width = value
            elif code == 43:
                self.vertical_height = value
            elif code == 44:
                self.line_spacing = value
            elif code == 50:
                rotation = value
            elif code == 71:
                self.attachment_point = value
            else:
                yield code, value  # chain of generators

        lines.append(text)
        self.raw_text = "".join(lines)
        if xdir is None:
            xdir = deg2vec(rotation)
        self.xdirection = Vector(xdir).normalized()
        self.set_default_extrusion()

    def lines(self):
        return self.raw_text.split('\P')

    def plain_text(self, split=False):
        chars = []
        raw_chars = list(reversed(self.raw_text))  # text splitted into chars, in reversed order for efficient pop()
        while len(raw_chars):
            char = raw_chars.pop()
            if char == '\\':  # is a formatting command
                try:
                    char = raw_chars.pop()
                except IndexError:
                    break  # premature end of text - just ignore

                if char in ESCAPED_CHARS:  # \ { }
                    chars.append(char)
                elif char in ONE_CHAR_COMMANDS:
                    if char == 'P':  # new line
                        chars.append('\n')
                        # discard other commands
                else:  # more character commands are terminated by ';'
                    stacking = char == 'S'  # stacking command surrounds user data
                    try:
                        while char != ';':  # end of format marker
                            char = raw_chars.pop()
                            if stacking and char != ';':
                                chars.append(char)  # append user data of stacking command
                    except IndexError:
                        break  # premature end of text - just ignore
            elif char in GROUP_CHARS:  # { }
                pass  # discard group markers
            elif char == '%':  # special characters
                if len(raw_chars) and raw_chars[-1] == '%':
                    raw_chars.pop()  # discard next '%'
                    if len(raw_chars):
                        special_char = raw_chars.pop()
                        # replace or discard formatting code
                        chars.append(SPECIAL_CHARS.get(special_char, ""))
                else:  # char is just a single '%'
                    chars.append(char)
            else:  # char is what it is, a character
                chars.append(char)

        plain_text = "".join(chars)
        return plain_text.split('\n') if split else plain_text

    def resolve_text_style(self, text_styles):
        style = text_styles.get(self.style, None)
        if style is None:
            style = default_text_style
        if self.height == 0:
            self.height = style.height
        if self.font is None:
            self.font = style.font
        if self.big_font is None:
            self.big_font = style.font


class Light(DXFEntity):
    def __init__(self):
        super(Light, self).__init__()
        self.version = 1
        self.name = ""
        self.light_type = 1  # distant = 1; point = 2; spot = 3
        self.status = False  # on/off ?
        self.light_color = 0  # 0 is unset
        self.true_color = None  # None is unset
        self.plot_glyph = 0
        self.intensity = 0
        self.position = (0., 0., 1.)
        self.target = (0., 0., 0.)
        self.attenuation_type = 0  # 0 = None; 1 = Inverse Linear; 2 = Inverse Square
        self.use_attenuation_limits = False
        self.attenuation_start_limit = 0
        self.attenuation_end_limit = 0
        self.hotspot_angle = 0
        self.fall_off_angle = 0.
        self.cast_shadows = False
        self.shadow_type = 0  # 0 = Ray traced shadows; 1 = Shadow maps
        self.shadow_map_size = 0
        self.shadow_softness = 0

    def setup_attributes(self, tags):
        for code, value in super(Light, self).setup_attributes(tags):
            if code == 1:
                self.name = value
            elif code == 10:
                self.position = value
            elif code == 11:
                self.target = value
            elif code == 40:
                self.intensity = value
            elif code == 41:
                self.attenuation_start_limit = value
            elif code == 42:
                self.attenuation_end_limit = value
            elif code == 50:
                self.hotspot_angle = value
            elif code == 51:
                self.fall_off_angle = value
            elif code == 63:
                self.light_color = value
            elif code == 70:
                self.light_type = value
            elif code == 72:
                self.attenuation_type = value
            elif code == 73:
                self.shadow_type = value
            elif code == 90:
                self.version = value
            elif code == 91:
                self.shadow_map_size = value
            elif code == 280:
                self.shadow_softness = value
            elif code == 290:
                self.status = value
            elif code == 291:
                self.plot_glyph = value
            elif code == 292:
                self.use_attenuation_limits = value
            elif code == 293:
                self.cast_shadows = value
            elif code == 421:
                self.true_color = value
            else:
                yield code, value  # chain of generators


class Body(DXFEntity):
    def __init__(self):
        super(Body, self).__init__()
        # need handle to get SAB data in DXF version AC1027 and later
        self.version = 1
        self.acis = []

    def setup_attributes(self, tags):
        sat = []
        for code, value in super(Body, self).setup_attributes(tags):
            if code == 70:
                self.version = value
            elif code in (1, 3):
                sat.append(value)
            else:
                yield code, value  # chain of generators
        self.acis = decode(sat)

    def set_sab_data(self, sab_data):
        self.acis = sab_data

    @property
    def is_sat(self):
        return isinstance(self.acis, list)  # but could be an empty list

    @property
    def is_sab(self):
        return not self.is_sat  # has binary encoded ACIS data


class Surface(Body):
    def __init__(self):
        super(Body, self).__init__()
        self.u_isolines = 0
        self.v_isolines = 0

    def setup_attributes(self, tags):
        for code, value in super(Surface, self).setup_attributes(tags):
            if code == 71:
                self.u_isolines = value
            elif code == 72:
                self.v_isolines = value
            else:
                yield code, value  # chain of generators


class Mesh(DXFEntity):
    def __init__(self):
        super(Mesh, self).__init__()
        self.version = 2
        self.blend_crease = False
        self.subdivision_levels = 1
        # rest are mostly positional tags
        self.vertices = []
        self.faces = []
        self.edges = []
        self.edge_crease_list = []

    def setup_attributes(self, tags):
        status = 0
        count = 0
        index_tags = []
        for code, value in super(Mesh, self).setup_attributes(tags):
            if code == 10:
                self.vertices.append(value)
            elif status == -1:  # ignore overridden properties at the end of the mesh entity
                pass  # property override uses also group codes 90, 91, 92 but only at the end of the MESH entity
            elif code == 71:
                self.version = value
            elif code == 72:
                self.blend_crease = bool(value)
            elif code == 91:
                self.subdivision_levels = value
            elif 92 <= code <= 95:  # 92 = vertices, 93 = faces; 94 = edges 95 = edge creases
                if code in (92, 95):
                    continue  # ignore vertices and edge creases count
                status = code
                count = value
                if status == 94:  # edge count
                    count *= 2
            elif code == 140:
                self.edge_crease_list.append(value)
            elif code == 90 and count > 0:  # faces or edges
                count -= 1
                index_tags.append(value)
                if count < 1:
                    if status == 93:
                        self.setup_faces(index_tags)
                    elif status == 94:
                        self.setup_edges(index_tags)
                    index_tags = []
            elif code == 90:  # count == 0; start of overridden properties (group code 90 after face or edge list)
                status = -1
            else:
                yield code, value  # chain of generators

    def get_face(self, index):
        return tuple(self.vertices[vertex_index] for vertex_index in self.faces[index])

    def get_edge(self, index):
        return tuple(self.vertices[vertex_index] for vertex_index in self.edges[index])

    def setup_faces(self, tags):
        face = []
        count = 0
        for value in tags:
            if count == 0:
                if len(face):
                    self.faces.append(tuple(face))
                    del face[:]
                count = value
            else:
                count -= 1
                face.append(value)

        if len(face):
            self.faces.append(tuple(face))

    def setup_edges(self, tags):
        self.edges = list(zip(tags[::2], tags[1::2]))


class Spline(DXFEntity):
    def __init__(self):
        super(Spline, self).__init__()
        self.normal_vector = None
        self.flags = 0
        self.degree = 3
        self.start_tangent = None
        self.end_tangent = None
        self.knots = []
        self.weights = []
        self.tol_knot = .0000001
        self.tol_control_point = .0000001
        self.tol_fit_point = .0000000001
        self.control_points = []
        self.fit_points = []

    def setup_attributes(self, tags):
        subclass = 'AcDbSpline'
        for code, value in super(Spline, self).setup_attributes(tags):
            if subclass == 'AcDbHelix':
                yield code, value # chain of generators
            elif code == 10:
                self.control_points.append(value)
            elif code == 11:
                self.fit_points.append(value)
            elif code == 12:
                self.start_tangent = value
            elif code == 13:
                self.end_tangent = value
            elif code == 40:
                self.knots.append(value)
            elif code == 41:
                self.weights.append(value)
            elif code == 42:
                self.tol_knot = value
            elif code == 43:
                self.tol_control_point = value
            elif code == 44:
                self.tol_fit_point = value
            elif code == 70:
                self.flags = value
            elif code == 71:
                self.degree = value
            elif 72 <= code < 75:
                pass  # ignore knot-, control- and fit point count
            elif code == 100:
                subclass = value
        self.normal_vector = self.extrusion
        if len(self.weights) == 0:
            self.weights = [1.0] * len(self.control_points)

    @property
    def is_closed(self):
        return bool(self.flags & const.SPLINE_CLOSED)

    @property
    def is_periodic(self):
        return bool(self.flags & const.SPLINE_PERIODIC)

    @property
    def is_rational(self):
        return bool(self.flags & const.SPLINE_RATIONAL)

    @property
    def is_planar(self):
        return bool(self.flags & const.SPLINE_PLANAR)

    @property
    def is_linear(self):
        return bool(self.flags & const.SPLINE_LINEAR)


class Helix(Spline):
    def __init__(self):
        super(Helix, self).__init__()
        self.helix_version = (1, 1)
        self.axis_base_point = None
        self.start_point = None
        self.axis_vector = None
        self.radius = 0
        self.turns = 0
        self.turn_height = 0
        self.handedness = 0  # 0 = left, 1 = right
        self.constrain = 0
        # 0 = Constrain turn height;
        # 1 = Constrain turns;
        # 2 = Constrain height

    def setup_attributes(self, tags):
        helix_major_version = 1
        helix_maintainance_version = 1
        for code, value in super(Helix, self).setup_attributes(tags):
            if code == 10:
                self.axis_base_point = value
            elif code == 11:
                self.start_point = value
            elif code == 12:
                self.axis_vector = value
            elif code == 90:
                helix_major_version = value
            elif code == 91:
                helix_maintainance_version = value
            elif code == 91:
                helix_maintainance_version = value
            elif code == 40:
                self.radius = value
            elif code == 41:
                self.turns = value
            elif code == 42:
                self.turn_height = value
            elif code == 290:
                self.handedness = value
            elif code == 280:
                self.constrain = value
            else:
                yield code, value  # chain of generators
        self.helix_version = (helix_major_version, helix_maintainance_version)


EntityTable = {
    'LINE': Line,
    'POINT': Point,
    'CIRCLE': Circle,
    'ARC': Arc,
    'TRACE': Trace,
    'SOLID': Solid,
    '3DFACE': Face,
    'TEXT': Text,
    'INSERT': Insert,
    'SEQEND': DXFEntity,
    'ATTRIB': Attrib,
    'ATTDEF': Attrib,
    'POLYLINE': Polyline,
    'VERTEX': Vertex,
    'BLOCK': Block,
    'ENDBLK': DXFEntity,
    'LWPOLYLINE': LWPolyline,
    'ELLIPSE': Ellipse,
    'RAY': Ray,
    'XLINE': Ray,
    'SPLINE': Spline,
    'HELIX': Helix,
    'MTEXT': MText,
    'MESH': Mesh,
    'LIGHT': Light,
    'BODY': Body,
    'REGION': Body,
    '3DSOLID': Body,
    'SURFACE': Surface,
    'PLANESURFACE': Surface,
}


def entity_factory(tags):
    dxftype = tags.get_type()
    cls = EntityTable[dxftype]  # get entity class or raise KeyError
    entity = cls()  # call constructor
    list(entity.setup_attributes(tags))  # setup dxf attributes - chain of generators
    return entity



# SPDX-FileCopyrightText: 2019-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy
from bpy.types import Panel

from .operators import get_rig_and_cam, CameraRigMixin


class ADD_CAMERA_RIGS_PT_camera_rig_ui(Panel, CameraRigMixin):
    bl_label = "Camera Rig"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Item'

    def draw(self, context):
        active_object = context.active_object
        rig, cam = get_rig_and_cam(context.active_object)
        pose_bones = rig.pose.bones
        cam_data = cam.data
        layout = self.layout

        # Camera lens
        if rig["rig_id"].lower() in ("dolly_rig", "crane_rig"):
            layout.prop(pose_bones["Camera"], '["lens"]',
                        text="Focal Length (mm)")

        col = layout.column(align=True)
        col.label(text="Clipping:")
        col.prop(cam_data, "clip_start", text="Start")
        col.prop(cam_data, "clip_end", text="End")

        layout.prop(cam_data, "type")

        # DoF
        col = layout.column(align=False)
        col.prop(cam_data.dof, "use_dof")
        if cam_data.dof.use_dof:
            sub = col.column(align=True)
            if cam_data.dof.focus_object is None:
                sub.operator("add_camera_rigs.set_dof_bone")
            sub.prop(cam_data.dof, "focus_object")
            if (cam_data.dof.focus_object is not None
                    and cam_data.dof.focus_object.type == 'ARMATURE'):
                sub.prop_search(cam_data.dof, "focus_subtarget",
                                cam_data.dof.focus_object.data, "bones")
            sub = col.column(align=True)
            row = sub.row(align=True)
            row.active = cam_data.dof.focus_object is None
            row.prop(pose_bones["Camera"],
                     '["focus_distance"]', text="Focus Distance")
            sub.prop(pose_bones["Camera"],
                     '["aperture_fstop"]', text="F-Stop")

        # Viewport display
        layout.prop(active_object, 'show_in_front',
                    toggle=False, text='Show in Front')
        layout.prop(cam_data, "show_limits")
        col = layout.column(align=True)
        col.prop(cam_data, "show_passepartout")
        if cam_data.show_passepartout:
            col.prop(cam_data, "passepartout_alpha")

        # Composition guides
        layout.popover(
            panel="ADD_CAMERA_RIGS_PT_composition_guides",
            text="Composition Guides",)

        # Props and operators
        col = layout.column(align=True)
        col.prop(cam,
                    "hide_select", text="Make Camera Unselectable")
        col.operator("add_camera_rigs.add_marker_bind",
                     text="Add Marker and Bind", icon="MARKER_HLT")
        col.operator("add_camera_rigs.set_scene_camera",
                     text="Make Camera Active", icon='CAMERA_DATA')

        if rig["rig_id"].lower() in ("dolly_rig", "crane_rig"):
            # Track to Constraint
            col = layout.column(align=True)
            track_to_constraint = None
            for con in pose_bones["Camera"].constraints:
                if con.type == 'TRACK_TO':
                    track_to_constraint = con
                    break
            if track_to_constraint is not None:
                col.label(text="Tracking:")
                col.prop(track_to_constraint, 'influence',
                         text="Aim Lock", slider=True)

            # Crane arm stuff
            if rig["rig_id"].lower() == "crane_rig":
                col = layout.column(align=True)
                col.label(text="Crane Arm:")
                col.prop(pose_bones["Crane_height"],
                         'scale', index=1, text="Arm Height")
                col.prop(pose_bones["Crane_arm"],
                         'scale', index=1, text="Arm Length")

        # 2D rig stuff
        elif rig["rig_id"].lower() == "2d_rig":
            col = layout.column(align=True)
            col.label(text="2D Rig:")
            col.prop(pose_bones["Camera"], '["rotation_shift"]',
                     text="Rotation/Shift")
            if cam.data.sensor_width != 36:
                col.label(text="Please set Camera Sensor Width to 36", icon="ERROR")


def register():
    bpy.utils.register_class(ADD_CAMERA_RIGS_PT_camera_rig_ui)


def unregister():
    bpy.utils.unregister_class(ADD_CAMERA_RIGS_PT_camera_rig_ui)



# SPDX-FileCopyrightText: 2019-2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy
from bpy.types import (
        Operator,
        Menu,
        )
from bpy.props import (
        BoolProperty,
        StringProperty,
        )

from bl_ui.properties_paint_common import UnifiedPaintPanel



# Animation Player (Thanks to marvin.k.breuer) #
class VIEW3D_MT_Animation_Player(Menu):
    bl_label = "Animation"

    def draw(self, context):
        layout = self.layout

        layout.operator("screen.animation_play", text="PLAY", icon='PLAY')
        layout.operator("screen.animation_play", text="Stop", icon='PAUSE')
        layout.operator("screen.animation_play", text="Reverse", icon='PLAY_REVERSE').reverse = True
        layout.separator()


        layout.operator("screen.keyframe_jump", text="Next FR", icon='NEXT_KEYFRAME').next = True
        layout.operator("screen.keyframe_jump", text="Previous FR", icon='PREV_KEYFRAME').next = False
        layout.separator()

        layout.operator("screen.frame_jump", text="Jump FF", icon='FF').end = True
        layout.operator("screen.frame_jump", text="Jump REW", icon='REW').end = False
        layout.separator()

        layout.menu("VIEW3D_MT_object_animation", text="Keyframes", icon='DECORATE_ANIMATE')



# List The Classes #

classes = (
    VIEW3D_MT_Animation_Player,
)


# Register Classes & Hotkeys #
def register():
    for cls in classes:
        bpy.utils.register_class(cls)


# Unregister Classes & Hotkeys #
def unregister():

    for cls in reversed(classes):
        bpy.utils.unregister_class(cls)


if __name__ == "__main__":
    register()



# SPDX-FileCopyrightText: 2010-2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

bl_info = {
    "name": "Scatter Objects",
    "author": "Jacques Lucke",
    "version": (0, 2),
    "blender": (3, 0, 0),
    "location": "3D View",
    "description": "Distribute object instances on another object.",
    "warning": "",
    "doc_url": "{BLENDER_MANUAL_URL}/addons/object/scatter_objects.html",
    "tracker_url": "https://projects.blender.org/blender/blender-addons/issues",
    "support": 'OFFICIAL',
    "category": "Object",
}

from . import ui
from . import operator

def register():
    ui.register()
    operator.register()

def unregister():
    ui.unregister()
    operator.unregister()



# SPDX-FileCopyrightText: 2022-2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

from typing import List, Sequence, Tuple

import bpy
import functools
import os
from bpy.app.handlers import persistent

from bpy.types import (
    Context,
    MetaSequence,
    Operator,
    PropertyGroup,
    SceneSequence,
    Window,
    WindowManager,
)
from bpy.props import (
    BoolProperty,
    IntProperty,
    StringProperty,
)
from .scene_tools import STORYPENCIL_OT_NewScene
from .render import STORYPENCIL_OT_RenderAction
from .sound import delete_sounds, send_sound_to_strip_scene

def window_id(window: Window) -> str:
    """ Get Window's ID.

    :param window: the Window to consider
    :return: the Window's ID
    """
    return str(window.as_pointer())


def get_window_from_id(wm: WindowManager, win_id: str) -> Window:
    """Get a Window object from its ID (serialized ptr).

    :param wm: a WindowManager holding Windows
    :param win_id: the ID of the Window to get
    :return: the Window matching the given ID, None otherwise
    """
    return next((w for w in wm.windows if w and window_id(w) == win_id), None)


def get_main_windows_list(wm: WindowManager) -> Sequence[Window]:
    """Get all the Main Windows held by the given WindowManager `wm`"""
    return [w for w in wm.windows if w and w.parent is None]


def join_win_ids(ids: List[str]) -> str:
    """Join Windows IDs in a single string"""
    return ";".join(ids)


def split_win_ids(ids: str) -> List[str]:
    """Split a Windows IDs string into individual IDs"""
    return ids.split(";")


class STORYPENCIL_OT_SetSyncMainOperator(Operator):
    bl_idname = "storypencil.sync_set_main"
    bl_label = "Set as Sync Main"
    bl_description = "Set this Window as main for Synchronization"
    bl_options = {'INTERNAL'}

    win_id: bpy.props.StringProperty(
        name="Window ID",
        default="",
        options=set(),
        description="Main window ID",
    )

    def copy_settings(self, main_window, secondary_window):
        if main_window is None or secondary_window is None:
            return
        secondary_window.scene.storypencil_main_workspace = main_window.scene.storypencil_main_workspace
        secondary_window.scene.storypencil_main_scene = main_window.scene.storypencil_main_scene
        secondary_window.scene.storypencil_edit_workspace = main_window.scene.storypencil_edit_workspace

    def execute(self, context):
        options = context.window_manager.storypencil_settings
        options.main_window_id = self.win_id
        wm = bpy.context.window_manager
        scene = context.scene
        wm['storypencil_use_new_window'] = scene.storypencil_use_new_window

        main_windows = get_main_windows_list(wm)
        main_window = get_main_window(wm)
        secondary_window = get_secondary_window(wm)
        # Active sync
        options.active = True
        if secondary_window is None:
            # Open a new window
            if len(main_windows) < 2:
                bpy.ops.storypencil.create_secondary_window()
                secondary_window = get_secondary_window(wm)
                self.copy_settings(get_main_window(wm), secondary_window)
                return {'FINISHED'}
            else:
                # Reuse the existing window
                secondary_window = get_not_main_window(wm)
        else:
            # Open new secondary
            if len(main_windows) < 2:
                bpy.ops.storypencil.create_secondary_window()
                secondary_window = get_secondary_window(wm)
                self.copy_settings(get_main_window(wm), secondary_window)
                return {'FINISHED'}
            else:
                # Reuse the existing window
                secondary_window = get_not_main_window(wm)

        if secondary_window:
            enable_secondary_window(wm, window_id(secondary_window))
            win_id = window_id(secondary_window)
            self.copy_settings(get_main_window(wm), secondary_window)
            bpy.ops.storypencil.sync_window_bring_front(win_id=win_id)

        return {'FINISHED'}


class STORYPENCIL_OT_AddSecondaryWindowOperator(Operator):
    bl_idname = "storypencil.create_secondary_window"
    bl_label = "Create Secondary Window"
    bl_description = "Create a Secondary Main Window and enable Synchronization"
    bl_options = {'INTERNAL'}

    def execute(self, context):
        # store existing windows
        windows = set(context.window_manager.windows[:])
        bpy.ops.wm.window_new_main()
        # get newly created window by comparing to previous list
        new_window = (set(context.window_manager.windows[:]) - windows).pop()
        # activate sync system and enable sync for this window
        toggle_secondary_window(context.window_manager, window_id(new_window))
        context.window_manager.storypencil_settings.active = True
        # trigger initial synchronization to open the current Sequence's Scene
        on_frame_changed()
        # Configure the new window
        self.configure_new_secondary_window(context, new_window)

        return {'FINISHED'}

    def configure_new_secondary_window(self, context, new_window):
        wrk_name = context.scene.storypencil_edit_workspace.name
        # Open the 2D workspace
        blendpath = os.path.dirname(bpy.app.binary_path)
        version = bpy.app.version
        version_full = str(version[0]) + '.' + str(version[1])
        template = os.path.join("scripts", "startup",
                                "bl_app_templates_system")
        template = os.path.join(template, wrk_name, "startup.blend")
        template_path = os.path.join(blendpath, version_full, template)
        # Check if workspace exist and add it if missing
        for wk in bpy.data.workspaces:
            if wk.name == wrk_name:
                new_window.workspace = wk
                return
        with context.temp_override(window=new_window):
            bpy.ops.workspace.append_activate(idname=wrk_name, filepath=template_path)


class STORYPENCIL_OT_WindowBringFront(Operator):
    bl_idname = "storypencil.sync_window_bring_front"
    bl_label = "Bring Window Front"
    bl_description = "Bring a Window to Front"
    bl_options = {'INTERNAL'}

    win_id: bpy.props.StringProperty()

    def execute(self, context):
        win = get_window_from_id(context.window_manager, self.win_id)
        if not win:
            return {'CANCELLED'}
        with context.temp_override(window=win):
            bpy.ops.wm.window_fullscreen_toggle()
            bpy.ops.wm.window_fullscreen_toggle()
        return {'FINISHED'}


class STORYPENCIL_OT_WindowCloseOperator(Operator):
    bl_idname = "storypencil.close_secondary_window"
    bl_label = "Close Window"
    bl_description = "Close a specific Window"
    bl_options = {'INTERNAL'}

    win_id: bpy.props.StringProperty()

    def execute(self, context):
        win = get_window_from_id(context.window_manager, self.win_id)
        if not win:
            return {'CANCELLED'}
        with context.temp_override(window=win):
            bpy.ops.wm.window_close()
        return {'FINISHED'}


def validate_sync(window_manager: WindowManager) -> bool:
    """
    Ensure synchronization system is functional, with a valid main window.
    Disable it otherwise and return the system status.
    """
    if not window_manager.storypencil_settings.active:
        return False
    if not get_window_from_id(window_manager, window_manager.storypencil_settings.main_window_id):
        window_manager.storypencil_settings.active = False
    return window_manager.storypencil_settings.active


def get_secondary_window_indices(wm: WindowManager) -> List[str]:
    """Get secondary Windows indices as a list of IDs

    :param wm: the WindowManager to consider
    :return: the list of secondary Windows IDs
    """
    return split_win_ids(wm.storypencil_settings.secondary_windows_ids)


def is_secondary_window(window_manager: WindowManager, win_id: str) -> bool:
    """Return wether the Window identified by 'win_id' is a secondary window.

    :return: whether this Window is a sync secondary
    """
    return win_id in get_secondary_window_indices(window_manager)


def enable_secondary_window(wm: WindowManager, win_id: str):
    """Enable the secondary status of a Window.

    :param wm: the WindowManager instance
    :param win_id: the id of the window
    """
    secondary_indices = get_secondary_window_indices(wm)
    win_id_str = win_id
    # Delete old indice if exist
    if win_id_str in secondary_indices:
        secondary_indices.remove(win_id_str)

    # Add indice
    secondary_indices.append(win_id_str)

    # rebuild the whole list of valid secondary windows
    secondary_indices = [
        idx for idx in secondary_indices if get_window_from_id(wm, idx)]

    wm.storypencil_settings.secondary_windows_ids = join_win_ids(secondary_indices)


def toggle_secondary_window(wm: WindowManager, win_id: str):
    """Toggle the secondary status of a Window.

    :param wm: the WindowManager instance
    :param win_id: the id of the window
    """
    secondary_indices = get_secondary_window_indices(wm)
    win_id_str = win_id
    if win_id_str in secondary_indices:
        secondary_indices.remove(win_id_str)
    else:
        secondary_indices.append(win_id_str)

    # rebuild the whole list of valid secondary windows
    secondary_indices = [
        idx for idx in secondary_indices if get_window_from_id(wm, idx)]

    wm.storypencil_settings.secondary_windows_ids = join_win_ids(secondary_indices)


def get_main_window(wm: WindowManager) -> Window:
    """Get the Window used to drive the synchronization system

    :param wm: the WindowManager instance
    :returns: the main Window or None
    """
    return get_window_from_id(wm=wm, win_id=wm.storypencil_settings.main_window_id)


def get_secondary_window(wm: WindowManager) -> Window:
    """Get the first secondary Window

    :param wm: the WindowManager instance
    :returns: the Window or None
    """
    for w in wm.windows:
        win_id = window_id(w)
        if is_secondary_window(wm, win_id):
            return w

    return None


def get_not_main_window(wm: WindowManager) -> Window:
    """Get the first not main Window

    :param wm: the WindowManager instance
    :returns: the Window or None
    """
    for w in wm.windows:
        win_id = window_id(w)
        if win_id != wm.storypencil_settings.main_window_id:
            return w

    return None


def get_main_strip(wm: WindowManager) -> SceneSequence:
    """Get Scene Strip at current time in Main window

    :param wm: the WindowManager instance
    :returns: the Strip at current time or None
    """
    context = bpy.context
    scene = context.scene
    main_scene = scene.storypencil_main_scene

    if main_scene.storypencil_use_new_window:
        main_window = get_main_window(wm=wm)
        if not main_window or not main_window.scene.sequence_editor:
            return None
        seq_editor = main_window.scene.sequence_editor
        return seq_editor.sequences.get(wm.storypencil_settings.main_strip_name, None)
    else:
        seq_editor = main_scene.sequence_editor.sequences
        for strip in seq_editor:
            if strip.type != 'SCENE':
                continue
            if strip.scene.name == scene.name:
                return strip

    return None


class STORYPENCIL_OT_SyncToggleSecondary(Operator):
    bl_idname = "storypencil.sync_toggle_secondary"
    bl_label = "Toggle Secondary Window Status"
    bl_description = "Enable/Disable synchronization for a specific Window"
    bl_options = {'INTERNAL'}

    win_id: bpy.props.StringProperty(name="Window Index")

    def execute(self, context):
        wm = context.window_manager
        toggle_secondary_window(wm, self.win_id)
        return {'FINISHED'}


def get_sequences_at_frame(
        frame: int,
        sequences: Sequence[Sequence]) -> Sequence[bpy.types.Sequence]:
    """ Get all sequencer strips at given frame.

    :param frame: the frame to consider
    """
    return [s for s in sequences if frame >= s.frame_start + s.frame_offset_start and
                                    frame < s.frame_start + s.frame_offset_start + s.frame_final_duration and
                                    s.type == 'SCENE']


def get_sequence_at_frame(
        frame: int,
        sequences: Sequence[bpy.types.Sequence] = None,
        skip_muted: bool = True,
) -> Tuple[bpy.types.Sequence, int]:
    """
    Get the higher sequence strip in channels stack at current frame.
    Recursively enters scene sequences and returns the original frame in the
    returned strip's time referential.

    :param frame: the frame to consider
    :param skip_muted: skip muted strips
    :returns: the sequence strip and the frame in strip's time referential
    """

    strips = get_sequences_at_frame(frame, sequences or bpy.context.sequences)

    # exclude muted strips
    if skip_muted:
        strips = [strip for strip in strips if not strip.mute]

    if not strips:
        return None, frame

    # Remove strip not scene type. Switch is only with Scenes
    for strip in strips:
        if strip.type != 'SCENE':
            strips.remove(strip)

    # consider higher strip in stack
    strip = sorted(strips, key=lambda x: x.channel)[-1]
    # go deeper when current strip is a MetaSequence
    if isinstance(strip, MetaSequence):
        return get_sequence_at_frame(frame, strip.sequences, skip_muted)
    if isinstance(strip, SceneSequence):
        # apply time offset to get in sequence's referential
        frame = frame - strip.frame_start + strip.scene.frame_start
        # enter scene's sequencer if used as input
        if strip.scene_input == 'SEQUENCER':
            return get_sequence_at_frame(frame, strip.scene.sequence_editor.sequences)
    return strip, frame


def set_scene_frame(scene, frame, force_update_main=False):
    """
    Set `scene` frame_current to `frame` if different.

    :param scene: the scene to update
    :param frame: the frame value
    :param force_update_main: whether to force the update of main scene
    """
    options = bpy.context.window_manager.storypencil_settings
    if scene.frame_current != frame:
        scene.frame_current = int(frame)
        scene.frame_set(int(frame))
        if force_update_main:
            update_sync(
                bpy.context, bpy.context.window_manager.storypencil_settings.main_window_id)


def setup_window_from_scene_strip(window: Window, strip: SceneSequence):
    """Change the Scene and camera of `window` based on `strip`.

    :param window: [description]
    :param scene_strip: [description]
    """
    if window.scene != strip.scene:
        window.scene = strip.scene
    if strip.scene_camera and strip.scene_camera != window.scene.camera:
        strip.scene.camera = strip.scene_camera


@persistent
def on_frame_changed(*args):
    """
    React to current frame changes and synchronize secondary windows.
    """
    # ensure context is fully initialized, i.e not '_RestrictData
    if not isinstance(bpy.context, Context):
        return

    # happens in some cases (not sure why)
    if not bpy.context.window:
        return

    wm = bpy.context.window_manager

    # early return if synchro is disabled / not available
    if not validate_sync(wm) or len(bpy.data.scenes) < 2:
        return

    # get current window id
    update_sync(bpy.context)


def update_sync(context: Context, win_id=None):
    """ Update synchronized Windows based on the current `context`.

    :param context: the context
    :param win_id: specify a window id (context.window is used otherwise)
    """
    wm = context.window_manager

    if not win_id:
        win_id = window_id(context.window)

    main_scene = get_window_from_id(
        wm, wm.storypencil_settings.main_window_id).scene
    if not main_scene.sequence_editor:
        return

    # return if scene's sequence editor has no sequences
    sequences = main_scene.sequence_editor.sequences
    if not sequences:
        return

    # bidirectionnal sync: change main time from secondary window
    if (
            win_id != wm.storypencil_settings.main_window_id
            and is_secondary_window(wm, win_id)
    ):
        # get strip under time cursor in main window
        strip, old_frame = get_sequence_at_frame(
            main_scene.frame_current,
            sequences=sequences
        )
        # only do bidirectional sync if secondary window matches the strip at current time in main
        if not isinstance(strip, SceneSequence) or strip.scene != context.scene:
            return

        # calculate offset
        frame_offset = context.scene.frame_current - old_frame
        if frame_offset == 0:
            return

        new_main_frame = main_scene.frame_current + frame_offset
        update_main_time = True
        # check if a valid scene strip is available under new frame before changing main time
        f_start = strip.frame_start + strip.frame_offset_start
        f_end = f_start + strip.frame_final_duration
        if new_main_frame < f_start or new_main_frame >= f_end:
            new_strip, _ = get_sequence_at_frame(
                new_main_frame,
                main_scene.sequence_editor.sequences,
            )
            update_main_time = isinstance(new_strip, SceneSequence)
        if update_main_time:
            # update main time change in the next event loop + force the sync system update
            # because Blender won't trigger a frame_changed event to avoid infinite recursion
            bpy.app.timers.register(
                functools.partial(set_scene_frame, main_scene,
                                  new_main_frame, True)
            )

        return

    # return if current window is not main window
    if win_id != wm.storypencil_settings.main_window_id:
        return

    secondary_windows = [
        get_window_from_id(wm, win_id)
        for win_id
        in get_secondary_window_indices(wm)
        if win_id and win_id != wm.storypencil_settings.main_window_id
    ]

    # only work with at least 2 windows
    if not secondary_windows:
        return

    seq, frame = get_sequence_at_frame(main_scene.frame_current, sequences)

    # return if no sequence at current time or not a scene strip
    if not isinstance(seq, SceneSequence) or not seq.scene:
        wm.storypencil_settings.main_strip_name = ""
        return

    wm.storypencil_settings.main_strip_name = seq.name
    # change the scene on secondary windows
    # warning: only one window's scene can be changed in this event loop,
    #          otherwise it may crashes Blender randomly
    for idx, win in enumerate(secondary_windows):
        if not win:
            continue
        # change first secondary window immediately
        if idx == 0:
            setup_window_from_scene_strip(win, seq)
        else:
            # trigger change in next event loop for other windows
            bpy.app.timers.register(
                functools.partial(setup_window_from_scene_strip, win, seq)
            )

    set_scene_frame(seq.scene, frame)


def sync_all_windows(wm: WindowManager):
    """Enable synchronization on all main windows held by `wm`."""
    wm.storypencil_settings.secondary_windows_ids = join_win_ids([
        window_id(w)
        for w
        in get_main_windows_list(wm)
    ])


@persistent
def sync_autoconfig(*args):
    """Autoconfigure synchronization system.
    If a window contains a VSE area on a scene with a valid sequence_editor,
    makes it main window and enable synchronization on all other main windows.
    """
    main_windows = get_main_windows_list(bpy.context.window_manager)
    # don't try to go any further if only one main window
    if len(main_windows) < 2:
        return

    # look for a main window with a valid sequence editor
    main = next(
        (
            win
            for win in main_windows
            if win.scene.sequence_editor
            and any(area.type == 'SEQUENCE_EDITOR' for area in win.screen.areas)
        ),
        None
    )
    # if any, set as main and activate sync on all other windows
    if main:
        bpy.context.window_manager.storypencil_settings.main_window_id = window_id(
            main)
        sync_all_windows(bpy.context.window_manager)
        bpy.context.window_manager.storypencil_settings.active = True


def sync_active_update(self, context):
    """ Update function for WindowManager.storypencil_settings.active. """
    # ensure main window is valid, using current context's window if none is set
    if (
            self.active
            and (
            not self.main_window_id
            or not get_window_from_id(context.window_manager, self.main_window_id)
            )
    ):
        self.main_window_id = window_id(context.window)
        # automatically sync all other windows if nothing was previously set
        if not self.secondary_windows_ids:
            sync_all_windows(context.window_manager)

    on_frame_changed()


def draw_sync_header(self, context):
    """Draw Window sync tools header."""

    wm = context.window_manager
    self.layout.separator()
    if wm.get('storypencil_use_new_window') is not None:
        new_window = wm['storypencil_use_new_window']
    else:
        new_window = False

    if not new_window:
        if context.scene.storypencil_main_workspace:
            if context.scene.storypencil_main_workspace.name != context.workspace.name:
                if context.area.ui_type == 'DOPESHEET':
                    row = self.layout.row(align=True)
                    row.operator(STORYPENCIL_OT_Switch.bl_idname,
                                 text="Back To VSE")


def draw_sync_sequencer_header(self, context):
    """Draw Window sync tools header."""
    if context.space_data.view_type != 'SEQUENCER':
        return

    wm = context.window_manager
    layout = self.layout
    layout.separator()
    row = layout.row(align=True)
    row.label(text="Scenes:")
    if context.scene.storypencil_use_new_window:
        row.operator(STORYPENCIL_OT_SetSyncMainOperator.bl_idname, text="Edit")
    else:
        row.operator(STORYPENCIL_OT_Switch.bl_idname, text="Edit")

    row.separator()
    layout.operator_context = 'INVOKE_REGION_WIN'
    row.operator(STORYPENCIL_OT_NewScene.bl_idname, text="New")

    layout.operator_context = 'INVOKE_DEFAULT'
    row.separator(factor=0.5)
    row.operator(STORYPENCIL_OT_RenderAction.bl_idname, text="Render")


class STORYPENCIL_PG_Settings(PropertyGroup):
    """
    PropertyGroup with storypencil settings.
    """
    active: BoolProperty(
        name="Synchronize",
        description=(
            "Automatically open current Sequence's Scene in other "
            "Main Windows and activate Time Synchronization"),
        default=False,
        update=sync_active_update
    )

    main_window_id: StringProperty(
        name="Main Window ID",
        description="ID of the window driving the Synchronization",
        default="",
    )

    secondary_windows_ids: StringProperty(
        name="Secondary Windows",
        description="Serialized Secondary Window Indices",
        default="",
    )

    active_window_index: IntProperty(
        name="Active Window Index",
        description="Index for using Window Manager's windows in a UIList",
        default=0
    )

    main_strip_name: StringProperty(
        name="Main Strip Name",
        description="Scene Strip at current time in the Main window",
        default="",
    )

    show_main_strip_range: BoolProperty(
        name="Show Main Strip Range in Secondary Windows",
        description="Draw main Strip's in/out markers in synchronized secondary Windows",
        default=True,
    )


# -------------------------------------------------------------
# Switch manually between Main and Edit Scene and Layout
#
# -------------------------------------------------------------
class STORYPENCIL_OT_Switch(Operator):
    bl_idname = "storypencil.switch"
    bl_label = "Switch"
    bl_description = "Switch workspace"
    bl_options = {'REGISTER', 'UNDO'}

    # Get active strip
    def act_strip(self, context):
        scene = context.scene
        sequences = scene.sequence_editor.sequences
        if not sequences:
            return None
        # Get strip under time cursor
        strip, old_frame = get_sequence_at_frame(
            scene.frame_current, sequences=sequences)
        return strip

    # ------------------------------
    # Poll
    # ------------------------------
    @classmethod
    def poll(cls, context):
        scene = context.scene
        if scene.storypencil_main_workspace is None or scene.storypencil_main_scene is None:
            return False
        if scene.storypencil_edit_workspace is None:
            return False

        return True

    # ------------------------------
    # Execute button action
    # ------------------------------
    def execute(self, context):
        wm = context.window_manager
        scene = context.scene
        wm['storypencil_use_new_window'] = scene.storypencil_use_new_window

        # Switch to Main
        if scene.storypencil_main_workspace.name != context.workspace.name:
            cfra_prv = scene.frame_current
            prv_pin = None
            if scene.storypencil_main_workspace is not None:
                if scene.storypencil_main_workspace.use_pin_scene:
                    scene.storypencil_main_workspace.use_pin_scene = False

                context.window.workspace = scene.storypencil_main_workspace

            if scene.storypencil_main_scene is not None:
                context.window.scene = scene.storypencil_main_scene
            strip = self.act_strip(context)
            if strip:
                context.window.scene.frame_current = int(cfra_prv + strip.frame_start) - 1

            # Delete sounds
            if scene.storypencil_copy_sounds:
                delete_sounds(scene)

            #bpy.ops.sequencer.reload()
        else:
            # Switch to Edit
            strip = self.act_strip(context)
            # save camera
            if strip is not None and strip.type == "SCENE":
                # Save data
                strip.scene.storypencil_main_workspace = scene.storypencil_main_workspace
                strip.scene.storypencil_main_scene = scene.storypencil_main_scene
                strip.scene.storypencil_edit_workspace = scene.storypencil_edit_workspace

                # Set workspace and Scene
                cfra_prv = scene.frame_current
                if scene.storypencil_edit_workspace.use_pin_scene:
                    scene.storypencil_edit_workspace.use_pin_scene = False

                context.window.workspace = scene.storypencil_edit_workspace
                context.window.workspace.update_tag()

                context.window.scene = strip.scene
                # Copy sounds
                if scene.storypencil_copy_sounds:
                    send_sound_to_strip_scene(strip, clear_sequencer=True, skip_mute=scene.storypencil_skip_sound_mute)


                active_frame = cfra_prv - strip.frame_start + 1
                if active_frame < strip.scene.frame_start:
                    active_frame = strip.scene.frame_start
                context.window.scene.frame_current = int(active_frame)

                # Set camera
                if strip.scene_input == 'CAMERA':
                    for screen in bpy.data.screens:
                        for area in screen.areas:
                            if area.type == 'VIEW_3D':
                                # select camera as view
                                if strip and strip.scene.camera is not None:
                                    area.spaces.active.region_3d.view_perspective = 'CAMERA'

                if scene.storypencil_copy_sounds:
                    bpy.ops.sequencer.reload()

        return {"FINISHED"}


class STORYPENCIL_OT_TabSwitch(Operator):
    bl_idname = "storypencil.tabswitch"
    bl_label = "Switch using tab key"
    bl_description = "Wrapper used to handle the Tab key to switch"
    bl_options = {'INTERNAL'}

    def execute(self, context):
        # For meta strips the tab key must be processed by other operator, so
        # just pass through to the next operator in the stack.
        if context.active_sequence_strip and context.active_sequence_strip.type == 'META':
            return {'PASS_THROUGH'}

        if context.scene.sequence_editor and context.scene.sequence_editor.meta_stack:
            return {'PASS_THROUGH'}

        if context.scene.storypencil_use_new_window:
            bpy.ops.storypencil.sync_set_main('INVOKE_DEFAULT', True)
        else:
            scene = context.scene
            sequences = scene.sequence_editor.sequences
            if sequences:
                # Get strip under time cursor
                strip, old_frame = get_sequence_at_frame(
                    scene.frame_current, sequences=sequences)
                if strip and strip.type == 'SCENE':
                    bpy.ops.storypencil.switch('INVOKE_DEFAULT', True)

        return {'FINISHED'}



# SPDX-FileCopyrightText: 2014-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later


class ArcEntity:
    """
    Used in convert.bulge_to_cubic() since bulges define how much a straight polyline segment should be transformed to
    an arc. ArcEntity is just used to call do.arc() without having a real Arc-Entity from dxfgrabber.
    """
    def __init__(self, start, end, center, radius, angdir):
        self.start_angle = start
        self.end_angle = end
        self.center = center
        self.radius = radius
        self.angdir = angdir

    def __str__(self):
        return "startangle: %s, endangle: %s, center: %s, radius: %s, angdir: %s" % \
               (str(self.start_angle), str(self.end_angle), str(self.center), str(self.radius), str(self.angdir))


class LineEntity:
    """
    Used in do._gen_meshface()
    """
    def __init__(self, start, end):
        self.start = start
        self.end = end
